/*
 * This file is (c) Copyright Zarafa, The Netherlands, 2007
 *
 * Use of this file and the source contained within it is subject to the
 * license of the package it belongs to. See the LICENSE file in the
 * top-level directory of the package.
 *
 */

#include "platform.h"

// Damn windows header defines max which break C++ header files
#undef max

#include <fstream>
#include <iostream>
#include <stdlib.h>
#include <sstream>

// vmime
#include "vmime/vmime.hpp"
#ifdef _WIN32
#include "vmime/platforms/windows/windowsHandler.hpp"
#else
#include "vmime/platforms/posix/posixHandler.hpp"
#endif

// mapi
#include <mapi.h>
#include <mapiutil.h>
#include <mapiext.h>
#include <edkmdb.h>

#include "tnef.h"

// inetmapi
#include "ECMapiUtils.h"
#include "MAPIToVMIME.h"
#include "VMIMEToMAPI.h"
#include "outputStreamMAPIAdapter.h"
#include "inputStreamMAPIAdapter.h"
#include "mapiAttachment.h"
#include "rtfutil.h"
#include "CommonUtil.h"
#include "Util.h"
#include "ECLogger.h"
#include "codepage.h"
#include "ecversion.h"

// icalmapi
#include "icalmapi.h"

// others

#ifdef _WIN32
#define BASETMPDIR "c:\\"
#else
#define BASETMPDIR "/tmp/"
#endif

#define SIZETOREAD 4096

#define SEND_ICAL 0

/* These are the properties that we DON'T send through TNEF. Reasons can be:
 *
 * - Privacy (PR_MDB_PROVIDER, etc)
 * - Sent through rfc2822 (PR_SUBJECT, PR_BODY, etc)
 * - Not part of an outgoing message (PR_TRANSPORT_MESSAGE_HEADERS)
 */

SizedSPropTagArray(1, sptaRTFInclude) = {1, {PR_RTF_COMPRESSED} };

SizedSPropTagArray(47, sptaExclude) = {
    47, 	{
        PR_BODY,
        PR_RTF_COMPRESSED,
        PR_RTF_IN_SYNC,
        PR_ACCESS,
        PR_ACCESS_LEVEL,
        PR_CLIENT_SUBMIT_TIME,
        PR_CREATION_TIME,
        PR_DELETE_AFTER_SUBMIT,
        PR_DISPLAY_BCC,
        PR_DISPLAY_CC,
        PR_DISPLAY_TO,
        PR_ENTRYID,
        PR_HASATTACH,
        PR_LAST_MODIFICATION_TIME,
        PR_MAPPING_SIGNATURE,
        PR_MDB_PROVIDER,
        PR_MESSAGE_DELIVERY_TIME,
        PR_MESSAGE_FLAGS,
        PR_MESSAGE_SIZE,
        PR_NORMALIZED_SUBJECT,
        PR_OBJECT_TYPE,
        PR_PARENT_ENTRYID,
        PR_RECORD_KEY,
        PR_RTF_SYNC_BODY_COUNT,
        PR_RTF_SYNC_BODY_CRC,
        PR_RTF_SYNC_BODY_TAG,
        PR_RTF_SYNC_PREFIX_COUNT,
        PR_RTF_SYNC_TRAILING_COUNT,
        PR_SEARCH_KEY,
        PR_SENDER_ADDRTYPE,
        PR_SENDER_EMAIL_ADDRESS,
        PR_SENDER_ENTRYID,
        PR_SENDER_NAME,
        PR_SENDER_SEARCH_KEY,
        PR_SENTMAIL_ENTRYID,
        PR_SENT_REPRESENTING_ADDRTYPE,
        PR_SENT_REPRESENTING_EMAIL_ADDRESS,
        PR_SENT_REPRESENTING_ENTRYID,
        PR_SENT_REPRESENTING_NAME,
        PR_SENT_REPRESENTING_SEARCH_KEY,
        PR_STORE_ENTRYID,
        PR_STORE_RECORD_KEY,
        PR_STORE_SUPPORT_MASK,
        PR_SUBJECT,
        PR_SUBJECT_PREFIX,
        PR_SUBMIT_FLAGS,
        PR_TRANSPORT_MESSAGE_HEADERS
    }
};

// These are named properties that are mapped to fixed values in Zarafa, so
// we can use the ID values instead of using GetIDsFromNames every times
#define PR_EC_USE_TNEF		PROP_TAG(PT_BOOLEAN, 0x8222)
#define PR_EC_SEND_AS_ICAL 	PROP_TAG(PT_BOOLEAN, 0x8000)
#define PR_EC_OUTLOOK_VERSION PROP_TAG(PT_STRING8, 0x81F4)

MAPIToVMIME::MAPIToVMIME()
{
	srand((unsigned)time(NULL));
	lpLogger = new ECLogger_File(EC_LOGLEVEL_ANY, 1, "-");	// TODO: remove fixed defaults
	delete_logger = true;
	m_lpAdrBook = NULL;
}

MAPIToVMIME::MAPIToVMIME(IMAPISession *lpSession, ECLogger *newlogger, sending_options sopt)
{
	srand((unsigned)time(NULL));
	lpLogger = newlogger;
	if (!lpLogger) {
		lpLogger = new ECLogger_File(EC_LOGLEVEL_ANY, 1, "-");	// TODO: remove fixed defaults
		delete_logger = true;
	} else {
		delete_logger = false;
	}
	
	this->sopt = sopt;
	if (lpSession) {
		lpSession->OpenAddressBook(0, NULL, AB_NO_DIALOG, &m_lpAdrBook);
		// ignore error
	} else {
		m_lpAdrBook = NULL;
	}
}

MAPIToVMIME::~MAPIToVMIME()
{
	if (delete_logger)
		delete lpLogger;

	if (m_lpAdrBook)
		m_lpAdrBook->Release();
}

HRESULT MAPIToVMIME::processRecipients(IMessage* lpMessage, vmime::messageBuilder* VMMessageBuilder, vmime::charset charset) {
	HRESULT			hr					= hrSuccess;
	vmime::mailbox	*lpVMMailbox		= NULL;
	LPMAPITABLE		lpRecipientTable	= NULL;
	LPSRowSet		pRows				= NULL;
	LPSPropValue	pPropRecipType		= NULL;
	LPSPropValue	pPropDispl			= NULL;
	LPSPropValue	pPropAType			= NULL;
	LPSPropValue	pPropEAddr			= NULL;
	bool			fToFound			= false;
	bool			hasRecips			= false;
	
	SizedSPropTagArray(6, sPropRecipColumns) = {6, { PR_ENTRYID, PR_EMAIL_ADDRESS, PR_DISPLAY_NAME, PR_RECIPIENT_TYPE, PR_SMTP_ADDRESS, PR_ADDRTYPE} };

	hr = lpMessage->GetRecipientTable(MAPI_DEFERRED_ERRORS, &lpRecipientTable);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to open recipient table. Error: 0x%08X", hr);
		goto exit;
	}

	hr = lpRecipientTable->SetColumns((LPSPropTagArray)&sPropRecipColumns, 0);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to set columns on recipient table. Error: 0x%08X", hr);
		goto exit;
	}

	hr = HrQueryAllRows(lpRecipientTable, NULL, NULL, NULL, 0, &pRows);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to read recipient table. Error: 0x%08X", hr);
		goto exit;
	}

	try {
		for (ULONG i = 0; i < pRows->cRows; i++) {
			pPropRecipType = PpropFindProp(pRows->aRow[i].lpProps, pRows->aRow[i].cValues, PR_RECIPIENT_TYPE);

			if(pPropRecipType == NULL) {
				// getMailBox properties
				pPropDispl = PpropFindProp(pRows->aRow[i].lpProps, pRows->aRow[i].cValues, PR_DISPLAY_NAME);
				pPropAType = PpropFindProp(pRows->aRow[i].lpProps, pRows->aRow[i].cValues, PR_ADDRTYPE);
				pPropEAddr = PpropFindProp(pRows->aRow[i].lpProps, pRows->aRow[i].cValues, PR_EMAIL_ADDRESS);
				lpLogger->Log(EC_LOGLEVEL_FATAL, "No recipient type set for recipient. DisplayName: %s, AddrType: %s, Email: %s",
							  pPropDispl?pPropDispl->Value.lpszA:"(none)",
							  pPropAType?pPropAType->Value.lpszA:"(none)",
							  pPropEAddr?pPropEAddr->Value.lpszA:"(none)");
				continue;
			}

			// ATTN: lpVMMailbox will be added to the VMMessageBuilder. The VMMessageBuilder will delete() it
			hr = MAPIToVMIME::getMailBox(&pRows->aRow[i], &lpVMMailbox, MAPI_CHARSET);
			if (hr != hrSuccess)
				goto exit;

			switch (pPropRecipType->Value.ul) {
			case MAPI_TO:
				VMMessageBuilder->getRecipients().appendAddress(lpVMMailbox);
				fToFound = true;
				break;
			case MAPI_CC:
				VMMessageBuilder->getCopyRecipients().appendAddress(lpVMMailbox);
				break;
			case MAPI_BCC:
				VMMessageBuilder->getBlindCopyRecipients().appendAddress(lpVMMailbox);
				break;
			}
			hasRecips = true;
		}
		
		if(!fToFound) {
			if (!hasRecips && sopt.no_recipients_workaround == false) {
				// spooler will exit here, gateway will continue with the workaround
				lpLogger->Log(EC_LOGLEVEL_FATAL, "No valid recipients found.");
				hr = MAPI_E_NOT_FOUND;
				goto exit;
			}

			// No recipients were in the 'To' .. add 'undisclosed recipients'
			vmime::mailboxGroup *undisclosed = new vmime::mailboxGroup(vmime::text("undisclosed-recipients"));
			VMMessageBuilder->getRecipients().appendAddress(undisclosed);
		}
			
	}
	catch (vmime::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception: "+e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (std::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on recipients: %s", e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (...) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception occurred on recipients");
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

exit:
	if (hr != hrSuccess)
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Could not read recipients");

	if (lpRecipientTable)
		lpRecipientTable->Release();

	if (pRows)
		FreeProws(pRows);

	return hr;
}

HRESULT MAPIToVMIME::handleSingleAttachment(IMessage* lpMessage, LPSRow lpRow, vmime::messageBuilder* lpVMMessageBuilder) {
	HRESULT			hr					= hrSuccess;
	IStream*		lpStream			= NULL; 
	LPATTACH		lpAttach			= NULL;
	LPSPropValue	pPropAttachNum		= NULL;
	LPSPropValue	pPropAttachType		= NULL;
	LPSPropValue	lpContentId 		= NULL;
	LPSPropValue	lpFilename			= NULL;
	ULONG			ulAttachmentNum		= 0;
	IMessage*		lpAttachedMessage	= NULL;
	inputStreamMAPIAdapter *inputDataStream = NULL;
	mapiAttachment *lpMapiAttach 		= NULL;
	vmime::defaultAttachment* lpVMDefAtt = NULL;
	std::string		strContentId		= "";
	sending_options sopt_keep;
	LPSPropValue	lpAMClass = NULL;
	LPSPropValue	lpAMAttach = NULL;
	LPSPropValue	lpMIMETag = NULL;
	char *			szFilename = NULL;  // just a reference, don't free
	vmime::mediaType szMIMEType;

	pPropAttachNum = PpropFindProp(lpRow->lpProps, lpRow->cValues, PR_ATTACH_NUM);
	if (pPropAttachNum == NULL) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Attachment in table not correct, no attachment number present.");
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	ulAttachmentNum = pPropAttachNum->Value.ul;

	// check PR_ATTACH_METHOD to determine Attachment or email
	pPropAttachType	= PpropFindProp(lpRow->lpProps, lpRow->cValues, PR_ATTACH_METHOD); 
	if (pPropAttachType == NULL) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Attachment in table not correct, no attachment method present for attachment %d.", ulAttachmentNum);
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	if (pPropAttachType->Value.ul == ATTACH_EMBEDDED_MSG) {
		vmime::message* lpVMNewMess = NULL;
		std::string strBuff;
		vmime::utility::outputStreamStringAdapter mout(strBuff);

		hr = lpMessage->OpenAttach(ulAttachmentNum, NULL, MAPI_BEST_ACCESS, &lpAttach);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Could not open message attachment %d. Error: 0x%08X", ulAttachmentNum, hr);
			goto exit;
		}

		hr = lpAttach->OpenProperty(PR_ATTACH_DATA_OBJ, &IID_IMessage, 0, MAPI_DEFERRED_ERRORS, (LPUNKNOWN *)&lpAttachedMessage);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Could not open data of message attachment %d. Error: 0x%08X", ulAttachmentNum, hr);
			goto exit;
		}

		// Check wether we're sending a iCal object
		// if so, we do not need to create a message-in-message object, but just attach the ics file.

		hr = HrGetOneProp(lpAttachedMessage, PR_MESSAGE_CLASS, &lpAMClass);
		if (hr == hrSuccess) {
			if (SEND_ICAL && strncmp(lpAMClass->Value.lpszA, "IPM.Note", 8) != 0) {
				// attach ical without submessage
				// still tnef when attachments are in the submessage
				hr = HrGetOneProp(lpAttachedMessage, PR_HASATTACH, &lpAMAttach);
				if (hr == hrSuccess && lpAMAttach->Value.b == false) {
					string ical;
					mapiAttachment *lpVMAttach = NULL;

					hr = HrMAPI2iCalendar(m_lpAdrBook, lpAttachedMessage, NULL, &ical);
					if (hr != hrSuccess) {
						lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to create ical object");
					} else {
						lpVMAttach = new mapiAttachment(vmime::stringContentHandler(ical),
														vmime::encoding("8bit"),
														vmime::mediaType("text/calendar"),
														"", vmime::word("calendar.ics"));
						// ics files are in utf-8
						lpVMAttach->addCharset(vmime::charset(vmime::charsets::UTF_8));
						lpVMMessageBuilder->appendAttachment(lpVMAttach);
						// and we're done
						goto exit;
					}
				}
			}
		}

		// sub objects do not necessarily need to have valid recipients, so disable the test
		sopt_keep = sopt;
		sopt.no_recipients_workaround = true;
		// recursive processing of embedded message as a new e-mail
		hr = MAPIToVMIME::convertMAPIToVMIME(lpAttachedMessage, &lpVMNewMess);
		if (hr != hrSuccess) {
			// Logging has been done by convertMAPIToVMIME()
			goto exit;
		}
		sopt = sopt_keep;

		lpVMNewMess->generate(mout);
		lpVMDefAtt = new vmime::defaultAttachment(vmime::stringContentHandler(strBuff),
												  vmime::encoding("8bit"),
												  vmime::mediaType("message/rfc822"));
		// append it (copies pointer, not data)
		lpVMMessageBuilder->appendAttachment(lpVMDefAtt);

	} else if (pPropAttachType->Value.ul == ATTACH_BY_VALUE) {
		hr = lpMessage->OpenAttach(ulAttachmentNum, NULL, MAPI_BEST_ACCESS, &lpAttach);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Could not open attachment %d. Error: 0x%08X", ulAttachmentNum, hr);
			goto exit;
		}
		
		if (HrGetOneProp(lpAttach, PR_ATTACH_CONTENT_ID, &lpContentId) == hrSuccess)
			strContentId = lpContentId->Value.lpszA;

		hr = lpAttach->OpenProperty(PR_ATTACH_DATA_BIN, &IID_IStream, 0, MAPI_DEFERRED_ERRORS, (LPUNKNOWN *)&lpStream);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Could not open data of attachment %d. Error: 0x%08X", ulAttachmentNum, hr);
			goto exit;
		}
	
		try {
			inputDataStream = new inputStreamMAPIAdapter(lpStream);	// add ref to lpStream
			
			
			// Set filename
			szFilename = "data.bin";
			
			if (HrGetOneProp(lpAttach, PR_ATTACH_LONG_FILENAME, &lpFilename) == hrSuccess) {
                szFilename = lpFilename->Value.lpszA;
            } else {
				if (HrGetOneProp(lpAttach, PR_ATTACH_FILENAME, &lpFilename) == hrSuccess) {
                    szFilename = lpFilename->Value.lpszA;
				}
            }
            
            // Set MIME type
            szMIMEType = parseMimeTypeFromFilename(szFilename);
            
            if (HrGetOneProp(lpAttach, PR_ATTACH_MIME_TAG, &lpMIMETag) == hrSuccess) {
                szMIMEType = lpMIMETag->Value.lpszA;
            }

			lpMapiAttach = new mapiAttachment(vmime::streamContentHandler(inputDataStream, 0 ,true),
											  vmime::encoding("base64"), szMIMEType,
											  strContentId, vmime::word(szFilename, MAPI_CHARSET));
											  
			// add to message (copies pointer, not data)
			lpVMMessageBuilder->appendAttachment(lpMapiAttach); 
		}
		catch (vmime::exception& e) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception: "+e.what());
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}
		catch (std::exception& e) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on attachment: %s", e.what());
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}
		catch (...) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception occurred on attachment");
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}
	} else {
		lpLogger->Log(EC_LOGLEVEL_WARNING, "Attachment %d contains invalid method %d. Not adding attachment to e-mail.", ulAttachmentNum, pPropAttachType->Value.ul);
	}

exit:
	// ATTN: lpMapiAttach are linked in the VMMessageBuilder. The VMMessageBuilder will delete() it.

	if (lpMIMETag)
        MAPIFreeBuffer(lpMIMETag);

	if (lpAMClass)
		MAPIFreeBuffer(lpAMClass);

	if (lpAMAttach)
		MAPIFreeBuffer(lpAMAttach);

	if (lpFilename) 
		MAPIFreeBuffer(lpFilename);

	if (lpContentId) 
		MAPIFreeBuffer(lpContentId);

	if (lpStream)
		lpStream->Release();

	if (lpAttach)
		lpAttach->Release();

	if (lpAttachedMessage)
		lpAttachedMessage->Release();

	return hr;
}

// TODO: FIXME with something nifty with /etc/mime.types
vmime::mediaType MAPIToVMIME::parseMimeTypeFromFilename(std::string strFilename) {
	std::string strExt;
	std::string strMedType;
	// to lowercase
	transform(strFilename.begin(), strFilename.end(), strFilename.begin(), ::tolower);
	strExt.assign(strFilename.substr(strFilename.find_last_of(".")+1, strFilename.length()));

	// application
	if (strExt == "bin" || strExt == "exe") {
		strMedType = "application/octet-stream";
	} else if (strExt == "ai" || strExt == "eps" || strExt == "ps") {
		strMedType = "application/postscript";
	} else if (strExt == "pdf") {
		strMedType = "application/pdf";
	} else if (strExt == "rtf") {
		strMedType = "application/rtf";
	} else if (strExt == "zip") {
		strMedType = "application/zip";
	} else if (strExt == "doc" || strExt == "dot") {
		strMedType = "application/msword";
	} else if (strExt == "mdb") {
		strMedType = "application/x-msaccess";
	} else if (strExt == "xla" || strExt == "xls" || strExt == "xlt" || strExt == "xlw") {
		strMedType = "application/vnd.ms-excel";
	} else if (strExt == "pot" || strExt == "ppt" || strExt == "pps") {
		strMedType = "application/vnd.ms-powerpoint";
	} else if (strExt == "mpp") {
		strMedType = "application/vnd.ms-project";
	}
	
	// audio

	else if (strExt == "ua") {
		strMedType = "audio/basic";
	} else if (strExt == "wav") {
		strMedType = "audio/x-wav";
	} else if (strExt == "mid") {
		strMedType = "audio/x-midi";
	}
	
	// image

	else if (strExt == "gif") {
		strMedType = "image/gif";
	} else if (strExt == "jpg" || strExt == "jpe" || strExt == "jpeg") {
		strMedType = "image/jpeg";
	} else if (strExt == "png") {
		strMedType = "image/png";
	} else if (strExt == "bmp") {
		strMedType = "image/x-ms-bmp";
	} else if (strExt == "tiff") {
		strMedType = "image/tiff";
	} else if (strExt == "xbm") {
		strMedType = "image/xbm";
	}
	
	// video

	else if (strExt == "mpg" || strExt == "mpe" || strExt == "mpeg") {
		strMedType = "video/mpeg";
	} else if (strExt == "qt" || strExt == "mov") {
		strMedType = "video/quicktime";
	} else if (strExt == "avi") {
		strMedType = "video/x-msvideo";
	} else {
		strMedType = "application/octet-stream";
	}

	return vmime::mediaType(strMedType);
}

HRESULT MAPIToVMIME::handleAttachments(IMessage* lpMessage, vmime::messageBuilder *lpVMMessageBuilder) {
	HRESULT		hr					= hrSuccess;
	LPSRowSet	pRows				= NULL;
	LPMAPITABLE	lpAttachmentTable	= NULL;

	// get attachment table
	hr = lpMessage->GetAttachmentTable(0, &lpAttachmentTable);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to open attachment table. Error: 0x%08X", hr);
		goto exit;
	}

	hr = HrQueryAllRows(lpAttachmentTable, NULL, NULL, NULL, 0, &pRows);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to fetch rows of attachment table. Error: 0x%08X", hr);
		goto exit;
	}
	
	// iterate through attachments and hand them off...
	for (ULONG i=0; i < pRows->cRows; i++) {
		hr = MAPIToVMIME::handleSingleAttachment(lpMessage, &pRows->aRow[i], lpVMMessageBuilder);
		// ignoring error... if one attachment fails we still want the others .. logging was done in handleSingleAttachment()
		hr = hrSuccess;
	} 

exit:
	if(pRows)
		FreeProws(pRows);

	if (lpAttachmentTable)
		lpAttachmentTable->Release();

	return hr;
}

HRESULT MAPIToVMIME::setBoundaries(vmime::header *lpVMHeader, vmime::body *lpVMBody, std::string boundary) {
    int i = 0;
    
    if(lpVMHeader->ContentType().getValue().getType() == "multipart") {
        // This is a multipart, so set the boundary for this part
        lpVMHeader->ContentType().setBoundary(boundary);
        
        // Set boundaries on children also
        for(i=0;i<lpVMBody->getPartCount();i++) {
            std::ostringstream os;

            vmime::bodyPart* VMBodyPart = lpVMBody->getPartAt(i);

            os << boundary << "_" << i;
            
            setBoundaries(VMBodyPart->getHeader(), VMBodyPart->getBody(), os.str());
        }
    }
    
    return hrSuccess;
}

HRESULT MAPIToVMIME::BuildNoteMessage(IMessage *lpMessage,  vmime::charset& charset, vmime::message **lppVMMessage)
{
	HRESULT					hr					= hrSuccess;
	LPSPropValue			lpRepresenting		= NULL;
	LPSPropValue			lpRepresentingEmail	= NULL;
	LPSPropValue			lpReplyTo			= NULL;
	LPSPropValue			lpMessageId			= NULL;
	LPSPropValue			lpReadReceipt		= NULL;
	LPSPropValue			lpImportance		= NULL;
	LPSPropValue			lpPriority			= NULL;
	LPSPropValue			lpSensitivity		= NULL;
	LPSPropValue			lpConversationIndex	= NULL;
	LPSPropValue			lpConversationTopic = NULL;
	LPSPropValue			lpNormSubject		= NULL;
	vmime::messageBuilder*	lpVMMessageBuilder	= NULL;
	vmime::message*			lpVMMessage			= NULL;	

	// construct the message..
	try {
		lpVMMessageBuilder = new vmime::messageBuilder();
		// messageBuilder is body and simple headers only (to/from/cc/subject/...)
		hr = MAPIToVMIME::fillVMIMEMail(lpMessage, lpVMMessageBuilder, charset);
		if (hr != hrSuccess)
			goto exit;			// Logging has been done in fillVMIMEMail()
		
		lpVMMessage = lpVMMessageBuilder->construct();
		// message builder has set Date header to now(). this will be overwritten.

		vmime::header* lpVMHeader = lpVMMessage->getHeader();

		// Conversation headers. New Message-Id header is set just before sending.		
		if(HrGetOneProp(lpMessage, PR_IN_REPLY_TO_ID, &lpMessageId) == hrSuccess) {
			// reply
			vmime::messageId* mid = new vmime::messageId(lpMessageId->Value.lpszA);
			lpVMHeader->InReplyTo().getValue().appendMessageId(mid);
			// vmime will delete the mid
			MAPIFreeBuffer(lpMessageId);
		}
		lpMessageId = NULL;

		// Outlook never adds this property
		if(HrGetOneProp(lpMessage, PR_INTERNET_REFERENCES, &lpMessageId) == hrSuccess) {
			// reply
			vmime::messageId* mid = new vmime::messageId(lpMessageId->Value.lpszA);
			lpVMHeader->References().getValue().appendMessageId(mid);
			// vmime will delete the mid
			MAPIFreeBuffer(lpMessageId);
		}
		lpMessageId = NULL;

		// priority settings
		char* priomap[3] = { "5 (Lowest)", "3 (Normal)", "1 (Highest)" }; // 2 and 4 cannot be set from outlook
		vmime::headerFieldFactory* hff = vmime::headerFieldFactory::getInstance();
		if (HrGetOneProp(lpMessage, PR_IMPORTANCE, &lpImportance) == hrSuccess) {
			lpVMHeader->appendField(hff->create("X-Priority", priomap[(lpImportance->Value.ul)&3])); // IMPORTANCE_* = 0..2
		} else if (HrGetOneProp(lpMessage, PR_PRIORITY, &lpPriority) == hrSuccess) {
			lpVMHeader->appendField(hff->create("X-Priority", priomap[(lpPriority->Value.ul+1)&3])); // PRIO_* = -1..1
		}

		// When adding a X-Priority, spamassassin may add a severe punishment because no User-Agent header
		// or X-Mailer header is present. So we set the X-Mailer header :)
		lpVMHeader->appendField(hff->create("X-Mailer", "Zarafa "PROJECT_VERSION_DOT_STR"-"PROJECT_SVN_REV_STR));

		// PR_CONVERSATION_INDEX
		if (HrGetOneProp(lpMessage, PR_CONVERSATION_INDEX, &lpConversationIndex) == hrSuccess) {
			vmime::string inString;
			inString.assign((const char*)lpConversationIndex->Value.bin.lpb, lpConversationIndex->Value.bin.cb);

			vmime::encoder* enc = vmime::encoderFactory::getInstance()->create("base64");
			vmime::utility::inputStreamStringAdapter in(inString);
			vmime::string outString;
			vmime::utility::outputStreamStringAdapter out(outString);

			enc->encode(in, out);
			delete enc;

			lpVMHeader->appendField(hff->create("Thread-Index", outString));
		}

		// PR_CONVERSATION_TOPIC is always the original started topic
		if (HrGetOneProp(lpMessage, PR_CONVERSATION_TOPIC, &lpConversationTopic) == hrSuccess) {
			if (HrGetOneProp(lpMessage, PR_NORMALIZED_SUBJECT, &lpNormSubject) != hrSuccess ||
				strcmp(lpNormSubject->Value.lpszA, lpConversationTopic->Value.lpszA) != 0)
			{
				removeEnters(lpConversationTopic->Value.lpszA);
				lpVMHeader->appendField(hff->create("Thread-Topic",
													vmime::text(lpConversationTopic->Value.lpszA, charset).generate()));
			}
		}

		if (HrGetOneProp(lpMessage, PR_SENSITIVITY, &lpSensitivity) == hrSuccess) {
			char *strSens;
			switch (lpSensitivity->Value.ul) {
			case SENSITIVITY_PERSONAL:
				strSens = "Personal";
				break;
			case SENSITIVITY_PRIVATE:
				strSens = "Private";
				break;
			case SENSITIVITY_COMPANY_CONFIDENTIAL:
				strSens = "Company-Confidential";
				break;
			case SENSITIVITY_NONE:
			default:
				strSens = NULL;
				break;
			};
			if (strSens)
				lpVMHeader->appendField(hff->create("Sensitivity", strSens));
		}

		// sender information
		std::string strEmAdd, strName, strType;
		std::string strResEmAdd, strResName, strResType;

		// Store owner, actual sender
		hr = HrGetAddress(m_lpAdrBook, lpMessage, PR_SENDER_ENTRYID, PR_SENDER_NAME, PR_SENDER_ADDRTYPE, PR_SENDER_EMAIL_ADDRESS, strName, strType, strEmAdd);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to get sender information. Error: 0x%08X", hr);
		    goto exit;
		}
		    
		hr = HrGetAddress(m_lpAdrBook, lpMessage, PR_SENT_REPRESENTING_ENTRYID, PR_SENT_REPRESENTING_NAME, PR_SENT_REPRESENTING_ADDRTYPE, PR_SENT_REPRESENTING_EMAIL_ADDRESS, strResName, strResType, strResEmAdd);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to get representing information. Error: 0x%08X", hr);
			goto exit;
		}

		// Set representing as from address, when possible
		// Ignore PR_SENT_REPRESENTING if the email adress is the same as the PR_SENDER email address
		if (strResEmAdd != "" && strResEmAdd != strEmAdd) {
			if (strResName == "" || strResName == strResEmAdd) 
				lpVMHeader->From().setValue(vmime::mailbox(strResEmAdd));
			else
				lpVMHeader->From().setValue(vmime::mailbox(vmime::text(strResName, MAPI_CHARSET), strResEmAdd));

			if(strResEmAdd != strEmAdd) {
				// Set store owner as sender
				if (strName == "" || strName == strEmAdd) 
					lpVMHeader->Sender().setValue(vmime::mailbox(strEmAdd));
				else
					lpVMHeader->Sender().setValue(vmime::mailbox(vmime::text(strName, MAPI_CHARSET), strEmAdd));
			}
		} else {
			// Set store owner as from, sender does not need to be set
			if (strName == "" || strName == strEmAdd) 
				lpVMHeader->From().setValue(vmime::mailbox(strEmAdd));
			else
				lpVMHeader->From().setValue(vmime::mailbox(vmime::text(strName, MAPI_CHARSET), strEmAdd));
		}

		// read receipt request
		if ((HrGetOneProp(lpMessage, PR_READ_RECEIPT_REQUESTED, &lpReadReceipt) == hrSuccess) && (lpReadReceipt->Value.b == TRUE)) {
			vmime::mailboxList mbl;
			if (strResEmAdd != "" && strResEmAdd != strEmAdd) {
				// use user added from address
				if (strResName == "" || strName == strResEmAdd)
					mbl.appendMailbox(new vmime::mailbox(strResEmAdd));
				else
					mbl.appendMailbox(new vmime::mailbox(vmime::text(strResName, MAPI_CHARSET), strResEmAdd));
			} else {
				if (strName == "" || strName == strEmAdd)
					mbl.appendMailbox(new vmime::mailbox(strEmAdd));
				else
					mbl.appendMailbox(new vmime::mailbox(vmime::text(strName, MAPI_CHARSET), strEmAdd));
			}
			lpVMHeader->DispositionNotificationTo().setValue(mbl);
		}
		
		// add reply to email
		if(HrGetOneProp(lpMessage, PR_REPLY_RECIPIENT_ENTRIES, &lpReplyTo) == hrSuccess) {
			if (lpReplyTo->Value.bin.cb > 0) {
				FLATENTRYLIST *lpEntryList = (FLATENTRYLIST *)lpReplyTo->Value.bin.lpb;

				if (lpEntryList->cEntries > 0) {
					FLATENTRY *lpEntry = (FLATENTRY *)&lpEntryList->abEntries;

					hr = HrGetAddress(m_lpAdrBook, (LPENTRYID)lpEntry->abEntry, lpEntry->cb, strName, strType, strEmAdd);
					if (hr == hrSuccess) {
						// vmime-0.7.1 can only set 1 email address in the ReplyTo field.
						if (strName != "" && strName != strEmAdd)
							lpVMHeader->ReplyTo().setValue(vmime::mailbox(vmime::text(strName, MAPI_CHARSET), strEmAdd));
						else
							lpVMHeader->ReplyTo().setValue(vmime::mailbox(strEmAdd));
					} else {
						// user selected a contact (or distrolist ?) in the reply-to
						// bounce conversion error back to the sender by leaving the hr result intact
						lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to set reply-to address");
						goto exit;
					}
				}
			}
		}

		// Set consistent boundary (optional)
		if(sopt.alternate_boundary) {
		    setBoundaries(lpVMMessage->getHeader(), lpVMMessage->getBody(), sopt.alternate_boundary);
		}

		// no hr checking
		handleXHeaders(lpMessage, lpVMMessage->getHeader());
	}
	catch (vmime::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception: "+e.what());
		hr = MAPI_E_CALL_FAILED; // set real error
		goto exit;
	}
	catch (std::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on note message: %s", e.what());
		hr = MAPI_E_CALL_FAILED; // set real error
		goto exit;
	}
	catch (...) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception on note message");
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	*lppVMMessage = lpVMMessage;

exit:
	
	if (lpRepresenting)
		MAPIFreeBuffer(lpRepresenting);
		
	if (lpRepresentingEmail)
		MAPIFreeBuffer(lpRepresentingEmail);
		
	if (lpReplyTo)
		MAPIFreeBuffer(lpReplyTo);

	if (lpReadReceipt)
		MAPIFreeBuffer(lpReadReceipt);

	if (lpImportance)
		MAPIFreeBuffer(lpImportance);

	if (lpPriority)
		MAPIFreeBuffer(lpPriority);

	if (lpSensitivity)
		MAPIFreeBuffer(lpSensitivity);

	if (lpConversationIndex)
		MAPIFreeBuffer(lpConversationIndex);

	if(lpConversationTopic)
		MAPIFreeBuffer(lpConversationTopic);

	if(lpNormSubject)
		MAPIFreeBuffer(lpNormSubject);

	if (lpVMMessageBuilder)
		delete(lpVMMessageBuilder);

	return hr;
}

HRESULT MAPIToVMIME::BuildMDNMessage(IMessage *lpMessage, vmime::charset& charset, vmime::message **lppVMMessage)
{
	HRESULT				hr = hrSuccess;
	vmime::message*		lpVMMessage = NULL;
	IStream*			lpBodyStream = NULL;
	LPSPropValue		lpiNetMsgId = NULL;
	LPSPropValue		lpMsgClass = NULL;
	LPSPropValue		lpSubject = NULL;
	LPMAPITABLE			lpRecipientTable	= NULL;
	LPSRowSet			pRows				= NULL;
	
	vmime::mailbox		expeditor; // From
	string				strMDNText;
	vmime::disposition	dispo;
	string				reportingUA; //empty
	std::vector<string>	reportingUAProducts; //empty
	vmime::message		msgOriginal;
	vmime::mailbox*		lpRecipientbox = NULL;
	string				strActionMode;
	
	// sender information
	std::string strEmailAdd, strName, strType;

	if (lpMessage == NULL || m_lpAdrBook == NULL || lppVMMessage == NULL) {
		hr = MAPI_E_INVALID_OBJECT;
		goto exit;
	}

	try {
		// Create original vmime message
		if (HrGetOneProp(lpMessage, PR_INTERNET_MESSAGE_ID, &lpiNetMsgId) == hrSuccess)
		{
			msgOriginal.getHeader()->OriginalMessageId().setValue(lpiNetMsgId->Value.lpszA);
			msgOriginal.getHeader()->MessageId().setValue(lpiNetMsgId->Value.lpszA);
		}
		
		// Create Recipient
		hr = lpMessage->GetRecipientTable(MAPI_DEFERRED_ERRORS, &lpRecipientTable);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to open MDN recipient table. Error: 0x%08X", hr);
			goto exit;
		}

		hr = HrQueryAllRows(lpRecipientTable, NULL, NULL, NULL, 0, &pRows);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to read MDN recipient table. Error: 0x%08X", hr);
			goto exit;
		}

		if (pRows->cRows == 0) {
			if (sopt.no_recipients_workaround == false) {
				lpLogger->Log(EC_LOGLEVEL_FATAL, "No MDN recipient found");
				hr = MAPI_E_NOT_FOUND;
				goto exit;
			} else {
				// no recipient, but need to continue ... is this correct??
				lpRecipientbox = new vmime::mailbox(string("undisclosed-recipients"));
			}
		}else {

			hr = MAPIToVMIME::getMailBox(&pRows->aRow[0], &lpRecipientbox, MAPI_CHARSET);
			if (hr != hrSuccess)
				goto exit;			// Logging done in getMailBox
		}

		vmime::mdn::sendableMDNInfos mdnInfos(&msgOriginal, *lpRecipientbox);
		
		//FIXME: Get the ActionMode and sending mode from the property 0x0081001E.
		//		And the type in property 0x0080001E.

		// Vmime is broken, The RFC 3798 says: "The action type can be 'manual-action' or 'automatic-action'"
		// but VMime can set only 'manual' or 'automatic' so should be add the string '-action'
		strActionMode = vmime::dispositionActionModes::MANUAL;
		strActionMode+= "-action";

		dispo.setActionMode(strActionMode);
		dispo.setSendingMode(vmime::dispositionSendingModes::SENT_MANUALLY);

		if (HrGetOneProp(lpMessage, PR_MESSAGE_CLASS, &lpMsgClass) != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "MDN message has no class.");
			hr = MAPI_E_CORRUPT_DATA;
			goto exit;
		}

		if(stricmp(lpMsgClass->Value.lpszA, "REPORT.IPM.Note.IPNNRN") == 0)
			dispo.setType(vmime::dispositionTypes::DELETED);
		else // if(stricmp(lpMsgClass->Value.lpszA, "REPORT.IPM.Note.IPNRN") == 0)
			dispo.setType(vmime::dispositionTypes::DISPLAYED);		
	
		strMDNText = "";// Default Empty body

		hr = lpMessage->OpenProperty(PR_BODY, &IID_IStream, 0, 0, (IUnknown**)&lpBodyStream);
		if(hr == hrSuccess)
		{

			hr = readStream(lpBodyStream, strMDNText);
			if (hr != hrSuccess) {
				lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to read MDN message body.");
				goto exit;
			}

		}

		// Store owner, actual sender
		hr = HrGetAddress(m_lpAdrBook, lpMessage, PR_SENDER_ENTRYID, PR_SENDER_NAME, PR_SENDER_ADDRTYPE, PR_SENDER_EMAIL_ADDRESS, strName, strType, strEmailAdd);
		if(hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to get MDN sender information. Error: 0x%08X", hr);
		    goto exit;
		}

		expeditor.setEmail(strEmailAdd);
		if(!strName.empty())
			expeditor.setName(vmime::text(strName, MAPI_CHARSET));
		else
			expeditor.setName(vmime::text(strEmailAdd, MAPI_CHARSET));

		//Create MDN message
		lpVMMessage = vmime::mdn::MDNHelper::buildMDN(mdnInfos, strMDNText, charset, expeditor, dispo, reportingUA, reportingUAProducts);
		if (!lpVMMessage) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to build MDN message");
			hr = MAPI_E_CORRUPT_DATA;
			goto exit;
		}

		// vmime is bad, so write subject
		if (HrGetOneProp(lpMessage, PR_SUBJECT, &lpSubject) == hrSuccess) {
			removeEnters(lpSubject->Value.lpszA);
			lpVMMessage->getHeader()->Subject().getValue() = (vmime::text)lpSubject->Value.lpszA;
		}

	}
	catch (vmime::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception: "+e.what());
	    hr = MAPI_E_CALL_FAILED; // set real error
	    goto exit;
	}
	catch (std::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on MDN message: %s", e.what());
	    hr = MAPI_E_CALL_FAILED; // set real error
	    goto exit;
	}
	catch (...) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception on MDN message");
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	*lppVMMessage = lpVMMessage;

exit:
	if(lpBodyStream)
		lpBodyStream->Release();

	if(lpRecipientbox)
		delete lpRecipientbox;

	if(lpiNetMsgId)
		MAPIFreeBuffer(lpiNetMsgId);
	
	if(lpMsgClass)
		MAPIFreeBuffer(lpMsgClass);

	if(lpRecipientTable)
		lpRecipientTable->Release();

	if(lpSubject)
		MAPIFreeBuffer(lpSubject);

	if(pRows)
		FreeProws(pRows);

	return hr;
}

HRESULT MAPIToVMIME::convertMAPIToVMIME(IMessage *lpMessage,  vmime::message **lppVMMessage) {
	HRESULT					hr					= hrSuccess;
	LPSPropValue			lpInternetCPID		= NULL;
	LPSPropValue			lpMsgClass			= NULL;
	vmime::message*			lpVMMessage			= NULL;
	vmime::charset			charset;
	char *					lpszCharset			= NULL;

	// Get the outgoing charset we want to be using
	if (HrGetOneProp(lpMessage, PR_INTERNET_CPID, &lpInternetCPID) == hrSuccess &&
		HrGetCharsetByCP(lpInternetCPID->Value.ul, &lpszCharset) == hrSuccess) {
		charset = vmime::charset(lpszCharset);
	} else {
		// default to windows-1252 if not set
		charset = vmime::charset("windows-1252");
	}

	if (HrGetOneProp(lpMessage, PR_MESSAGE_CLASS, &lpMsgClass) == hrSuccess &&
		(stricmp(lpMsgClass->Value.lpszA, "REPORT.IPM.Note.IPNNRN") == 0 ||
		stricmp(lpMsgClass->Value.lpszA, "REPORT.IPM.Note.IPNRN") == 0)
		)
	{
		// Create a read receipt message
		hr = BuildMDNMessage(lpMessage, charset, &lpVMMessage);
		if(hr != hrSuccess)
			goto exit;
	} else {
		// Create default
		hr = BuildNoteMessage(lpMessage, charset, &lpVMMessage);
		if (hr != hrSuccess)
			goto exit;
	}

	*lppVMMessage = lpVMMessage;

exit:
	if (lpInternetCPID)
		MAPIFreeBuffer(lpInternetCPID);

	if(lpMsgClass)
		MAPIFreeBuffer(lpMsgClass);

	return hr;
}


HRESULT MAPIToVMIME::fillVMIMEMail(IMessage *lpMessage, vmime::messageBuilder* lpVMMessageBuilder, vmime::charset charset) {
	std::string		strOut,	strOutConverted;
	HRESULT			hr				= hrSuccess;
	LPSRowSet		prows			= NULL;
	LPSPropValue	lpSubject		= NULL;
	LPSPropValue	lpUseTNEF		= NULL;
	LPSPropValue	lpSendAsICal	= NULL;
	LPSPropValue	lpOutlookVersion = NULL;
	LPSPropValue	lpMessageClass	= NULL;
	IStream 		*lpStream = NULL;
	BOOL			bRealRTF = FALSE;

	try {
		if (HrGetOneProp(lpMessage, PR_SUBJECT, &lpSubject) == hrSuccess) {
			removeEnters(lpSubject->Value.lpszA);
			strOut = lpSubject->Value.lpszA;
		}
		else
			strOut = "";

		vmime::charset::convert(strOut, strOutConverted, MAPI_CHARSET, charset);

		// we can't ignore any errors.. the sender should know if an email is not sent correctly
		lpVMMessageBuilder->setSubject(vmime::text(strOutConverted, charset));

		// handle recipients
		hr = MAPIToVMIME::processRecipients(lpMessage, lpVMMessageBuilder, charset);
		if (hr != hrSuccess)
			goto exit;			// Logging has been done in processRecipients()

		// handle text
		hr = MAPIToVMIME::handleTextparts(lpMessage, lpVMMessageBuilder, charset, &bRealRTF);
		if (hr != hrSuccess)
			goto exit;

		// handle attachments
		hr = MAPIToVMIME::handleAttachments(lpMessage, lpVMMessageBuilder);
		if (hr != hrSuccess)
			goto exit;

		// TNEF
		if(HrGetOneProp(lpMessage, PR_EC_USE_TNEF, &lpUseTNEF) != hrSuccess) {
			lpUseTNEF = NULL;
		}

		if(HrGetOneProp(lpMessage, PR_EC_SEND_AS_ICAL, &lpSendAsICal) != hrSuccess) {
			lpSendAsICal = NULL;
		}
		
		if(HrGetOneProp(lpMessage, PR_EC_OUTLOOK_VERSION, &lpOutlookVersion) != hrSuccess) {
			lpOutlookVersion = NULL;
		} 
		
		if(HrGetOneProp(lpMessage, PR_MESSAGE_CLASS, &lpMessageClass) != hrSuccess) {
			lpMessageClass = NULL;
		}

		if ((lpUseTNEF && lpUseTNEF->Value.b) || (lpSendAsICal && lpSendAsICal->Value.b) || 
			(lpSendAsICal && lpOutlookVersion && strcmp(lpOutlookVersion->Value.lpszA,"9.0") == 0) ||
			(lpMessageClass && (strncmp("IPM.Note", lpMessageClass->Value.lpszA, 8) != 0) ) || 
			bRealRTF == TRUE)
		{
			vmime::attachment *lpTNEFAtt;
			inputStreamMAPIAdapter *inputDataStream = NULL;

			lpLogger->Log(EC_LOGLEVEL_INFO, "Adding TNEF attachment for extra information");

			/* Send TNEF information for this message. This is done when Outlook requests
			 * this by setting PR_EC_USE_TNEF to TRUE or when PR_EC_SEND_AS_ICAL is TRUE. This is done
			 * because we currently do not support sending iCal messages, so we revert to TNEF when
			 * iCal is requested.
			 *
			 * Also, Outlook 2000 always sets PR_EC_SEND_AS_ICAL to FALSE, because the
			 * iCal option is somehow missing from the options property sheet, and 
			 * PR_EC_USE_TNEF is never set. So for outlook 2000 (9.0), we check the
			 * _existence_ of PR_EC_SEND_AS_ICAL as a hint to use TNEF.
			 *
			 * Also, anything that is *not* an IPM.Note is sent with TNEF data.
			 */
			 
			if (SEND_ICAL && lpMessageClass && (strncmp("IPM.Note", lpMessageClass->Value.lpszA, 8) != 0)) {
				// iCAL
				string ical;
				mapiAttachment *lpVMAttach = NULL;

				hr = HrMAPI2iCalendar(m_lpAdrBook, lpMessage, NULL, &ical);
				if (hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_WARNING, "Unable to create ical object, sending as TNEF");
					goto tnef_anyway;
				}

				lpVMAttach = new mapiAttachment(vmime::stringContentHandler(ical),
												vmime::encoding("8bit"),
												vmime::mediaType("text/calendar"),
												"", vmime::word("calendar.ics"));
				// ics files are in utf-8
				lpVMAttach->addCharset(vmime::charset(vmime::charsets::UTF_8));
				lpVMMessageBuilder->appendAttachment(lpVMAttach);
			} else {
tnef_anyway:
				hr = CreateStreamOnHGlobal(NULL, TRUE, &lpStream);
				if (hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to create stream for TNEF attachment. Error 0x%08X", hr);
					goto exit;
				}
			 
				ECTNEF tnef(TNEF_ENCODE, lpMessage, lpStream);
			
				// Encode the properties now
				hr = tnef.AddProps(TNEF_PROP_EXCLUDE, (LPSPropTagArray)&sptaExclude);
				if(hr != hrSuccess)
					goto exit;
			
				if (bRealRTF == TRUE) {
					hr = tnef.AddProps(TNEF_PROP_INCLUDE, (LPSPropTagArray)&sptaRTFInclude);
					if(hr != hrSuccess)
						goto exit;
				}
		
				// Write the stream
				hr = tnef.Finish();

				inputDataStream =  new inputStreamMAPIAdapter(lpStream);
			
				// Now, add the stream as an attachment to the message, filename winmail.dat 
				// and MIME type 'application/ms-tnef', no content-id
				lpTNEFAtt = new mapiAttachment(vmime::streamContentHandler(inputDataStream, 0 ,true),
											   vmime::encoding("base64"), vmime::mediaType("application/ms-tnef"), "",
											   vmime::word("winmail.dat"));
											  
				// add to message (copies pointer, not data)
				lpVMMessageBuilder->appendAttachment(lpTNEFAtt); 
			}
		}

		/*
		  To convert a VMMessageBuilder to a VMMessage object, an 'expeditor' needs to be set in vmime
		  later on, this from will be overwritten .. so maybe replace this with a dummy value?
		 */

		std::string strName, strType, strEmAdd;
		hr = HrGetAddress(m_lpAdrBook, lpMessage, PR_SENDER_ENTRYID, PR_SENDER_NAME, PR_SENDER_ADDRTYPE, PR_SENDER_EMAIL_ADDRESS, strName, strType, strEmAdd);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to get sender information. Error: 0x%08X", hr);
		    goto exit;
		}

		if (strName != "") {
			lpVMMessageBuilder->setExpeditor(vmime::mailbox(vmime::text(strName, MAPI_CHARSET), strEmAdd));
		} else {
			lpVMMessageBuilder->setExpeditor(vmime::mailbox(strEmAdd));
		}
		// sender and reply-to is set elsewhere because it can only be done on a vmime::message object...
	}
	catch (vmime::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception: "+e.what());
	    hr = MAPI_E_CALL_FAILED; // set real error
	    goto exit;
	}
	catch (std::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on fill message: %s", e.what());
	    hr = MAPI_E_CALL_FAILED; // set real error
	    goto exit;
	}
	catch (...) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception  on fill message");
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

exit:
	if (lpSubject)
		MAPIFreeBuffer(lpSubject);

	if (lpMessageClass)
		MAPIFreeBuffer(lpMessageClass);

	if (lpOutlookVersion)
		MAPIFreeBuffer(lpOutlookVersion);

	if (lpUseTNEF)
		MAPIFreeBuffer(lpUseTNEF);

	if (lpSendAsICal)
		MAPIFreeBuffer(lpSendAsICal);

	if (lpStream)
		lpStream->Release();

	if (prows)
		FreeProws(prows);

	return hr;
}


// make vmime::mailbox from row
HRESULT MAPIToVMIME::getMailBox(LPSRow lpRow, vmime::mailbox** lppVMMailbox, vmime::charset charset) {
	HRESULT			hr				= hrSuccess;
	vmime::mailbox	*lpVMMailbox	= NULL;
	std::string	strName, strEmail, strType;

	hr = HrGetAddress(m_lpAdrBook, lpRow->lpProps, lpRow->cValues, PR_ENTRYID, PR_DISPLAY_NAME, PR_ADDRTYPE, PR_EMAIL_ADDRESS, strName, strType, strEmail);
	if(hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_ERROR, "Unable to create mailbox. Error: %08X", hr);
		goto exit;
	}

	if (strName == "" || strName == strEmail) {
		lpVMMailbox = new vmime::mailbox(strEmail);
	} else {
		lpVMMailbox = new vmime::mailbox(vmime::text(strName, MAPI_CHARSET), strEmail);
	}

	*lppVMMailbox = lpVMMailbox;

exit:
	return hr;
}


// put textpart into PR_RTF_COMPRESSED
HRESULT MAPIToVMIME::handleTextparts(IMessage* lpMessage, vmime::messageBuilder* lpVMMessageBuilder, vmime::charset charset, BOOL* lpbRealRTF) {
	std::string strHTMLOut, strRtf;
	std::string strBody, strBodyConverted;
	HRESULT		hr						= hrSuccess;
	IStream*	lpCompressedRTFStream	= NULL; 
	IStream*	lpUncompressedRTFStream	= NULL; 
	char		*lpRBuff				= NULL;
	IStream*	lpBody					= NULL;
	IStream*	lpHTMLStream			= NULL;
	vmime::charset HTMLcharset;

	// set the encoder of plaintext body part
	vmime::encoding bodyEncoding("quoted-printable");
	// will skip enters in q-p encoding, "fixes" plaintext mails to be more plain text
	vmime::encoder *bodyEncoder = bodyEncoding.getEncoder();
	bodyEncoder->getProperties().setProperty("text", true);

	lpRBuff = new char[SIZETOREAD];

	if(lpbRealRTF) *lpbRealRTF = FALSE;

	// grabbing rtf
	hr = lpMessage->OpenProperty(PR_RTF_COMPRESSED, &IID_IStream, 0, 0, (LPUNKNOWN *)&lpCompressedRTFStream);
	if(hr == hrSuccess) {

		hr = WrapCompressedRTFStream(lpCompressedRTFStream, 0, &lpUncompressedRTFStream);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to create RTF-text stream. Error: 0x%08X", hr);
			goto exit;
		}

		hr = readStream(lpUncompressedRTFStream, strRtf);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to read RTF-text stream. Error: 0x%08X", hr);
			goto exit;
		}

		if(isrtfhtml(strRtf.c_str(), strRtf.size()) )
		{
			// Open html
			hr = lpMessage->OpenProperty(PR_HTML, &IID_IStream, 0, 0, (LPUNKNOWN *)&lpHTMLStream);
			if (hr == hrSuccess) {
				hr = readStream(lpHTMLStream, strHTMLOut);
				if (hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to read HTML-text stream. Error: 0x%08X", hr);
					goto exit;
				}

				// strHTMLout is now escaped us-ascii HTML, this is what we will be sending
				// Or, if something failed, the HTML is now empty
				lpHTMLStream->Release(); lpHTMLStream = NULL;
			} else {
				lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to open HTML-text stream. Error: 0x%08X", hr);
				// continue with plaintext
			}
		} else if(isrtftext(strRtf.c_str(), strRtf.size()) )	{
			//Do nothing, only plain/text
		} else {
			// Real rtf
			if(lpbRealRTF) *lpbRealRTF = TRUE;
		}
	} else {
		lpLogger->Log(EC_LOGLEVEL_INFO, "Unable to open rtf-text stream. Error: 0x%08X", hr);
		hr = hrSuccess;
	}
	
	try {
		hr = lpMessage->OpenProperty(PR_BODY, &IID_IStream, 0, 0, (IUnknown**)&lpBody);
		if (hr == hrSuccess) {
			hr = readStream(lpBody, strBody);
			if (hr != hrSuccess) {
				lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to read plain-text stream. Error: 0x%08X", hr);
				goto exit;
			}
			lpBody->Release(); lpBody = NULL;
		} else {
			lpLogger->Log(EC_LOGLEVEL_INFO, "Unable to open plain-text stream. Error: 0x%08X", hr);
			hr = hrSuccess;
		}

		// Convert body to correct charset		
		vmime::charset::convert(strBody, strBodyConverted, MAPI_CHARSET, charset);

		// If HTML, create HTML and plaintext parts
		if (!strHTMLOut.empty()) {
			lpVMMessageBuilder->constructTextPart(vmime::mediaType(vmime::mediaTypes::TEXT, vmime::mediaTypes::TEXT_HTML));
		 	vmime::htmlTextPart& textPart = dynamic_cast<vmime::htmlTextPart&>(*lpVMMessageBuilder->getTextPart());
			textPart.setText(vmime::stringContentHandler(strHTMLOut));
			textPart.setCharset(charset);
			if (strBodyConverted != "")
				textPart.setPlainText(vmime::stringContentHandler(strBodyConverted));
		} else { // else just plaintext
			if (strBodyConverted != "") {
				// make sure we give vmime CRLF data, so SMTP servers (qmail) won't complain on the forced plaintext
				char *crlfconv = NULL;
				size_t outsize = 0;
				crlfconv = new char[strBodyConverted.length()*2];
				Util::BufferLFtoCRLF(strBodyConverted.length(), strBodyConverted.c_str(), crlfconv, &outsize);
				strBodyConverted.assign(crlfconv, outsize);
				delete [] crlfconv;

				// encode to q-p ourselves
				vmime::utility::inputStreamStringAdapter in(strBodyConverted);
				vmime::string outString;
				vmime::utility::outputStreamStringAdapter out(outString);

				bodyEncoder->encode(in, out);

				lpVMMessageBuilder->getTextPart()->setCharset(charset);
				lpVMMessageBuilder->getTextPart()->setText(vmime::stringContentHandler(outString, bodyEncoding));
			}
		}
	}
	catch (vmime::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception: "+e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (std::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on text part: %s", e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (...) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception on text part");
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

exit:
	if (hr != hrSuccess)
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Could not parse mail body");

	if (lpHTMLStream)
		lpHTMLStream->Release();

	if (lpBody)
		lpBody->Release();
		
	if (lpRBuff)
		delete [] lpRBuff;

	if (lpUncompressedRTFStream)
		lpUncompressedRTFStream->Release();
		
	if (lpCompressedRTFStream)
		lpCompressedRTFStream->Release(); 

	if (bodyEncoder)
		delete bodyEncoder;

	return hr;
}

HRESULT MAPIToVMIME::handleXHeaders(IMessage *lpMessage, vmime::header *lpVMHeader) {
	HRESULT hr;
	ULONG i;
	ULONG cValues;
	LPSPropValue lpPropArray = NULL;
	LPSPropTagArray lpsPropTagArray = NULL;
	ULONG cNames;
	LPMAPINAMEID *lppNames = NULL;
	map<ULONG, ULONG> namemap;
	vmime::headerFieldFactory* hff = vmime::headerFieldFactory::getInstance();

	// get named props on message

	hr = lpMessage->GetProps(NULL, 0, &cValues, &lpPropArray);
	if (FAILED(hr))
		goto exit;

	cNames = 0;
	for (i = 0; i < cValues; i++) {
		if (PROP_ID(lpPropArray[i].ulPropTag) >= 0x8000)
			cNames++;
	}

	MAPIAllocateBuffer(CbNewSPropTagArray(cNames), (void **)&lpsPropTagArray);
	lpsPropTagArray->cValues = cNames;

	cNames = 0;
	for (i = 0; i < cValues; i++) {
		if (PROP_ID(lpPropArray[i].ulPropTag) >= 0x8000) {
			namemap[cNames] = i;
			lpsPropTagArray->aulPropTag[cNames++] = lpPropArray[i].ulPropTag;
		}
	}
	
	hr = lpMessage->GetNamesFromIDs(&lpsPropTagArray, NULL, 0, &cNames, &lppNames);
	if (FAILED(hr))
		goto exit;

	for (i = 0; i < cNames; i++) {
		if (lppNames[i] && lppNames[i]->ulKind == MNID_STRING && lppNames[i]->Kind.lpwstrName &&
			PROP_TYPE(lpPropArray[namemap[i]].ulPropTag) == PT_STRING8)
		{
			char *str;
			int l = unicodetombs(NULL, lppNames[i]->Kind.lpwstrName, 0) +1;
			str = new char[l];
			unicodetombs(str, lppNames[i]->Kind.lpwstrName, l);

			if ((str[0] == 'X' || str[0] == 'x') && str[1] == '-') {
				if (str[0] == 'x')
					capitalize(str);

				try {
					lpVMHeader->findField(str);
				}
				catch (vmime::exceptions::no_such_field) {
					lpVMHeader->appendField(hff->create(str, lpPropArray[namemap[i]].Value.lpszA));
				}
			}
			delete [] str;
		}
	}

exit:
	if (lpPropArray)
		MAPIFreeBuffer(lpPropArray);

	if (lpsPropTagArray)
		MAPIFreeBuffer(lpsPropTagArray);

	if (lppNames)
		MAPIFreeBuffer(lppNames);

	return hr;
}

// converts x-mail-sender -> X-Mail-Sender
void MAPIToVMIME::capitalize(char *s) {
	char *p;

	s[0] = toupper(s[0]);		// x to X
	p = s;
	while ((p = strchr(p, '-'))) { // capitalize every char after a -
		p++;
		if (*p != '\0')
			*p = toupper(*p);
	}
}

// makes spaces from enters, avoid spaces in Subject
void MAPIToVMIME::removeEnters(char *s) {
	char *p = s;

	while (*p) {
		if (*p == '\r' || *p == '\n') *p = ' ';
		p++;
	}
}
