/*
 * This file is (c) Copyright Zarafa, The Netherlands, 2007
 *
 * Use of this file and the source contained within it is subject to the
 * license of the package it belongs to. See the LICENSE file in the
 * top-level directory of the package.
 *
 */

#include "platform.h"

// Damn windows header defines max which break C++ header files
#undef max

#include "VMIMEToMAPI.h"
#include "ECGuid.h"

#include <string>
#include <fstream>
#include <iostream>
#include <stdlib.h>

// vmime
#include "vmime/vmime.hpp"
#ifdef _WIN32
#include "vmime/platforms/windows/windowsHandler.hpp"
#else
#include "vmime/platforms/posix/posixHandler.hpp"
#endif

// mapi
#include <mapi.h>
#include <mapix.h>
#include <mapiutil.h>
#include <mapiext.h>
#include <mapiguidext.h>
#include <edkmdb.h>

#include "EMSAbTag.h"
#include "tnef.h"
#include "codepage.h"
#include "icalmapi.h"
#include "Util.h"

// inetmapi
#include "ECMapiUtils.h"
#include "ECVMIMEUtils.h"

#include "outputStreamMAPIAdapter.h"

// rtf
#include <rtfutil.h>
#include <CommonUtil.h>

// if we can parse ical messages .. in development, so currently disabled
#define PARSE_ICAL 0

VMIMEToMAPI::VMIMEToMAPI()
{
	lpLogger = new ECLogger_File(EC_LOGLEVEL_ANY, 1, "-");	// TODO: remove fixed defaults
	delete_logger = true;
	imopt_default_delivery_options(&m_dopt);
	m_dopt.use_received_date = false; // use Date header
	m_lpAdrBook = NULL;
}

VMIMEToMAPI::VMIMEToMAPI(LPADRBOOK lpAdrBook, ECLogger *newlogger, delivery_options dopt)
{
	lpLogger = newlogger;
	m_dopt = dopt;
	if (!lpLogger) {
		lpLogger = new ECLogger_File(EC_LOGLEVEL_ANY, 1, "-");	// TODO: remove fixed defaults
		delete_logger = true;
	} else {
		delete_logger = false;
	}
	m_lpAdrBook = lpAdrBook;
}

VMIMEToMAPI::~VMIMEToMAPI()
{
	if (delete_logger)
		delete lpLogger;
}


HRESULT VMIMEToMAPI::convertVMIMEToMAPI(string input, IMessage *lpMessage) {
	HRESULT hr = hrSuccess;

	try {
		bodyLevel = BODY_NONE;

		// turn buffer into a message
		vmime::message VMMessage;
		VMMessage.parse(input);

		hr = fillMAPIMail(&VMMessage, lpMessage);
		if (hr != hrSuccess)
			goto exit;
	}
	catch (vmime::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception: "+e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (std::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception: %s", e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (...) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception occurred");
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

exit:
	return hr;
}

HRESULT VMIMEToMAPI::fillMAPIMail(vmime::message* VMMessage, IMessage *lpMessage) {
	HRESULT	hr = hrSuccess;
	SPropValue sPropDefaults[2];

	sPropDefaults[0].ulPropTag = PR_MESSAGE_CLASS;
	sPropDefaults[0].Value.lpszA = "IPM.Note";
	sPropDefaults[1].ulPropTag = PR_MESSAGE_FLAGS;
	sPropDefaults[1].Value.ul = m_dopt.mark_as_read ? MSGFLAG_READ : MSGFLAG_UNMODIFIED;

	hr = lpMessage->SetProps(2, sPropDefaults, NULL);
	if (hr != hrSuccess) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to set default mail properties");
		goto exit;
	}
	
	try {
		// turn buffer into a message

		// get the part header and find out what it is...
		vmime::header* VMHeader = VMMessage->getHeader();
		vmime::body* VMBody = VMMessage->getBody();

		// pass recipients somewhere else 
		hr = handleRecipients(VMHeader, lpMessage);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to parse mail recipients");
			goto exit;
		}

		// Headers
		hr = handleHeaders(VMHeader, lpMessage);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to parse mail headers");
			goto exit;
		}

		if(vmime::mdn::MDNHelper::isMDN(VMMessage) == true)
		{
			vmime::mdn::receivedMDNInfos receivedMDN = vmime::mdn::MDNHelper::getReceivedMDN(VMMessage);

			vmime::body* myBody = (vmime::body*)receivedMDN.getMessage()->getBody();
			// it is possible to get 3 bodyparts.
			// text/Plain, message/disposition-notification, text/rfc822-headers
			// Loop to get text/plain body.
			for (int i=0; i < myBody->getPartCount(); i++)
			{
				vmime::bodyPart* bPart = myBody->getPartAt(i);

				if(bPart->getHeader()->ContentType().getValue().getType() == vmime::mediaTypes::TEXT &&
					bPart->getHeader()->ContentType().getValue().getSubType() == vmime::mediaTypes::TEXT_PLAIN)
				{
					hr = disectBody(bPart->getHeader(), bPart->getBody(), lpMessage, true);
					if (hr != hrSuccess) {
						lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to parse MDN mail body");
						goto exit;
					}
				}
				
			}

			if(receivedMDN.getDisposition().getType() == vmime::dispositionTypes::DELETED)
			{
				sPropDefaults[0].ulPropTag = PR_MESSAGE_CLASS;
				sPropDefaults[0].Value.lpszA = "REPORT.IPM.Note.IPNNRN";				
			}else{
				sPropDefaults[0].ulPropTag = PR_MESSAGE_CLASS;
				sPropDefaults[0].Value.lpszA = "REPORT.IPM.Note.IPNRN";
			}

			string strId = "<"+receivedMDN.getOriginalMessageId().getId()+">";
			sPropDefaults[1].ulPropTag = 0x1046001E;//ptagOriginalInetMessageID
			sPropDefaults[1].Value.lpszA = (LPSTR)strId.c_str();

			hr = lpMessage->SetProps(2, sPropDefaults, NULL);
			if (hr != hrSuccess) {
				lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to set MDN mail properties");
				goto exit;
			}

		} else if (VMBody->getPartCount() > 0) {
			// multipart message, handle by part
			for (int i=0; i < VMBody->getPartCount(); i++) {
				vmime::bodyPart* VMBodyPart = VMBody->getPartAt(i);
				hr = disectBody(VMBodyPart->getHeader(), VMBodyPart->getBody(), lpMessage);
				if (hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to parse multipart %d of mail body", i);
					goto exit;
				}
			}
		} else {
			// one body part, always place in the body, no matter if the header says it's an _text_ attachment
			hr = disectBody(VMHeader, VMBody, lpMessage, true);
			if (hr != hrSuccess) {
				lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to parse mail body");
				goto exit;
			}
		}
	}
	catch (vmime::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception on create message: "+e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (std::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on create message: %s", e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (...) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception occurred on create message");
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

exit:
	return hr;
}

// put all headers where they belong..
HRESULT VMIMEToMAPI::handleHeaders(vmime::header* VMHeader, IMessage* lpMessage) {
	HRESULT			hr = hrSuccess;
	std::string		strHeaders, strInternetMessageId;
	std::string		strInReplyTos, strReferences;
	std::string		strSubject, strReplyTo;
	std::string		strClientTime;
	std::string		strFromEmail, strFromName, strFromSearchKey;;
	std::string		strSenderEmail,	strSenderName, strSenderSearchKey;
	ULONG			cbFromEntryID; // representing entry id
	LPENTRYID		lpFromEntryID = NULL;
	ULONG			cbSenderEntryID;
	LPENTRYID		lpSenderEntryID = NULL;
	LPSPropValue	lpPropNormalizedSubject = NULL;
	SPropValue		sConTopic;
	// setprops
	FLATENTRYLIST	*lpEntryList = NULL;
	int				nProps = 0;
	SPropValue		msgProps[20];
	// temp
	ULONG			cbEntryID;
	LPENTRYID		lpEntryID = NULL;

	LPSPropValue	lpRecipProps = NULL;
	ULONG			ulRecipProps;
	
	// order is important for modifyFromAddressBook()
	SizedSPropTagArray(7, sptaRecipPropsSentRepr) = { 7, {
		PR_SENT_REPRESENTING_ADDRTYPE, PR_SENT_REPRESENTING_NAME,
		PR_NULL /* PR_xxx_DISPLAY_TYPE not available */,
		PR_SENT_REPRESENTING_EMAIL_ADDRESS, PR_SENT_REPRESENTING_ENTRYID,
		PR_SENT_REPRESENTING_SEARCH_KEY, PR_NULL /* PR_xxx_SMTP_ADDRESS not available */
	} };
	SizedSPropTagArray(7, sptaRecipPropsSender) = { 7, {
		PR_SENDER_ADDRTYPE, PR_SENDER_NAME,
		PR_NULL /* PR_xxx_DISPLAY_TYPE not available */,
		PR_SENDER_EMAIL_ADDRESS, PR_SENDER_ENTRYID,
		PR_SENDER_SEARCH_KEY, PR_NULL /* PR_xxx_SMTP_ADDRESS not available */
	} };

	try { 
		int nIds, i;

		vmime::utility::outputStreamStringAdapter os(strHeaders);
		VMHeader->generate(os);
		msgProps[nProps].ulPropTag = PR_TRANSPORT_MESSAGE_HEADERS;
		msgProps[nProps++].Value.lpszA = (char*)strHeaders.c_str();

		// internet message ID
		strInternetMessageId = "<"+VMHeader->MessageId().getValue().getId()+">";
		msgProps[nProps].ulPropTag = PR_INTERNET_MESSAGE_ID;
		msgProps[nProps++].Value.lpszA = (char*)strInternetMessageId.c_str();

		// In-Reply-To header
		strInReplyTos.clear();
		nIds = VMHeader->InReplyTo().getValue().getMessageIdCount();
		for (i = 0; i < nIds; i++) {
		    // space/enter/nothing? between id's?
		    strInReplyTos += "<"+VMHeader->InReplyTo().getValue().getMessageIdAt(i)->getId()+"> ";
		}
		if (strInReplyTos.size() > 0) {
			msgProps[nProps].ulPropTag = PR_IN_REPLY_TO_ID;
			msgProps[nProps++].Value.lpszA = (char*)strInReplyTos.c_str();
		}

		// References header
		strReferences.clear();
		nIds = VMHeader->References().getValue().getMessageIdCount();
		for (i = 0; i < nIds; i++) {
		    strReferences += "<"+VMHeader->References().getValue().getMessageIdAt(i)->getId()+"> ";
		}
		if (strReferences.size() > 0) {
			msgProps[nProps].ulPropTag = PR_INTERNET_REFERENCES;
			msgProps[nProps++].Value.lpszA = (char*)strInReplyTos.c_str();
		}

		// set subject
		strSubject = VMHeader->Subject().getValue().getConvertedText(MAPI_CHARSET);
		msgProps[nProps].ulPropTag = PR_SUBJECT;
		msgProps[nProps++].Value.lpszA = (char*)strSubject.c_str();

		// set ReplyTo
		if (!VMHeader->ReplyTo().getValue().isEmpty()) {
			// First, set PR_REPLY_RECIPIENT_NAMES
			strReplyTo = VMHeader->ReplyTo().getValue().getName().getConvertedText(MAPI_CHARSET);
			if (strReplyTo == "")
				strReplyTo = VMHeader->ReplyTo().getValue().getEmail();

			msgProps[nProps].ulPropTag = PR_REPLY_RECIPIENT_NAMES;
			msgProps[nProps++].Value.lpszA = (char*)strReplyTo.c_str();

			// Now, set PR_REPLY_RECIPIENT_ENTRIES (a FLATENTRYLIST)
			hr = ECCreateOneOff((char *)strReplyTo.c_str(), "SMTP",
								(char *)VMHeader->ReplyTo().getValue().getEmail().c_str(), 0, &cbEntryID, &lpEntryID);
			if (hr != hrSuccess)
				goto exit;

			hr = MAPIAllocateBuffer(cbEntryID + sizeof(ULONG) + sizeof(FLATENTRYLIST), (void **)&lpEntryList);
			if (hr != hrSuccess)
				goto exit;

			memset(lpEntryList, 0, cbEntryID + sizeof(ULONG) + sizeof(FLATENTRYLIST));

			lpEntryList->cEntries = 1;
			lpEntryList->cbEntries = cbEntryID + sizeof(ULONG);
			*((ULONG *)&lpEntryList->abEntries) = cbEntryID;
			memcpy(&lpEntryList->abEntries+sizeof(ULONG), lpEntryID, cbEntryID);

			msgProps[nProps].ulPropTag = PR_REPLY_RECIPIENT_ENTRIES;
			msgProps[nProps].Value.bin.cb = cbEntryID + sizeof(ULONG) + sizeof(FLATENTRYLIST);
			msgProps[nProps++].Value.bin.lpb = (BYTE*)lpEntryList;

			MAPIFreeBuffer(lpEntryID);
			lpEntryID = NULL;
		}

		// setting sent date
		msgProps[nProps].ulPropTag = PR_CLIENT_SUBMIT_TIME;
		msgProps[nProps++].Value.ft = vmimeDatetimeToFiletime(VMHeader->Date().getValue());

		// setting receive date (now)
		msgProps[nProps].ulPropTag = PR_MESSAGE_DELIVERY_TIME;
		// parse from Received header, if possible. Otherwise, use now()
		if (m_dopt.use_received_date) {
			try {
				vmime::relayField *recv = dynamic_cast <vmime::relayField*>(VMHeader->findField("Received"));
				msgProps[nProps++].Value.ft = vmimeDatetimeToFiletime(recv->getValue().getDate());
			}
			catch (...) {
				GetSystemTimeAsFileTime(&msgProps[nProps++].Value.ft);
			}
		} else {
			GetSystemTimeAsFileTime(&msgProps[nProps++].Value.ft);
		}

		// The real sender of the mail
		strFromEmail = VMHeader->From().getValue().getEmail();
		if (!VMHeader->From().getValue().getName().isEmpty()) {
			strFromName = VMHeader->From().getValue().getName().getConvertedText(MAPI_CHARSET);
		} else {
			strFromName = strFromEmail;
		}

		hr = modifyFromAddressBook(&lpRecipProps, &ulRecipProps, strFromEmail.c_str(), MAPI_ORIG, (LPSPropTagArray)&sptaRecipPropsSentRepr);
		if (hr == hrSuccess) {
			hr = lpMessage->SetProps(ulRecipProps, lpRecipProps, NULL);
			if (hr != hrSuccess)
				goto exit;

			MAPIFreeBuffer(lpRecipProps);
			lpRecipProps = NULL;
		} else {
			msgProps[nProps].ulPropTag = PR_SENT_REPRESENTING_NAME;
			msgProps[nProps++].Value.lpszA = (char*)strFromName.c_str();

			msgProps[nProps].ulPropTag = PR_SENT_REPRESENTING_EMAIL_ADDRESS;
			msgProps[nProps++].Value.lpszA = (char*)strFromEmail.c_str();

			strFromSearchKey = "SMTP:"+strFromEmail;
			transform(strFromSearchKey.begin(), strFromSearchKey.end(), strFromSearchKey.begin(), ::toupper);
			msgProps[nProps].ulPropTag = PR_SENT_REPRESENTING_SEARCH_KEY;
			msgProps[nProps].Value.bin.cb = strFromSearchKey.size()+1; // include string terminator
			msgProps[nProps++].Value.bin.lpb = (BYTE*)strFromSearchKey.c_str();

			msgProps[nProps].ulPropTag = PR_SENT_REPRESENTING_ADDRTYPE;
			msgProps[nProps++].Value.lpszA = "SMTP";
			
			hr = ECCreateOneOff((char *)strFromName.c_str(), "SMTP", (char*)strFromEmail.c_str(), 0, &cbFromEntryID, &lpFromEntryID);
			if(hr != hrSuccess)
				goto exit;

			msgProps[nProps].ulPropTag = PR_SENT_REPRESENTING_ENTRYID;
			msgProps[nProps].Value.bin.cb = cbFromEntryID;
			msgProps[nProps++].Value.bin.lpb = (BYTE*)lpFromEntryID;

			// SetProps is later on...
		}

		// The original sender of the mail account (if non sender exist then the FROM)
		strSenderEmail = VMHeader->Sender().getValue().getEmail();
		if(VMHeader->Sender().getValue().getName().isEmpty() && strSenderEmail.empty()) {
			// Fallback on the original from address
			strSenderName = strFromName;
			strSenderEmail = strFromEmail;
		} else {
			if (!VMHeader->Sender().getValue().getName().isEmpty()) {
				strSenderName = VMHeader->Sender().getValue().getName().getConvertedText(MAPI_CHARSET);
			} else {
				strSenderName.assign(strFromEmail);
			}
		}

		hr = modifyFromAddressBook(&lpRecipProps, &ulRecipProps, strSenderEmail.c_str(), MAPI_ORIG, (LPSPropTagArray)&sptaRecipPropsSender);
		if (hr == hrSuccess) {
			hr = lpMessage->SetProps(ulRecipProps, lpRecipProps, NULL);
			if (hr != hrSuccess)
				goto exit;

			MAPIFreeBuffer(lpRecipProps);
			lpRecipProps = NULL;
		} else {
			msgProps[nProps].ulPropTag = PR_SENDER_NAME;
			msgProps[nProps++].Value.lpszA = (char*)strSenderName.c_str();

			msgProps[nProps].ulPropTag = PR_SENDER_EMAIL_ADDRESS;
			msgProps[nProps++].Value.lpszA = (char*)strSenderEmail.c_str();

			strSenderSearchKey = "SMTP:"+strSenderEmail;
			transform(strSenderSearchKey.begin(), strSenderSearchKey.end(), strSenderSearchKey.begin(), ::toupper);
			msgProps[nProps].ulPropTag = PR_SENDER_SEARCH_KEY;
			msgProps[nProps].Value.bin.cb = strSenderSearchKey.size()+1; // include string terminator
			msgProps[nProps++].Value.bin.lpb = (BYTE*)strSenderSearchKey.c_str();

			msgProps[nProps].ulPropTag = PR_SENDER_ADDRTYPE;
			msgProps[nProps++].Value.lpszA = "SMTP";
			
			hr = ECCreateOneOff((char *)strSenderName.c_str(), "SMTP", (char*)strSenderEmail.c_str(), 0, &cbSenderEntryID, &lpSenderEntryID);
			if(hr != hrSuccess)
				goto exit;

			msgProps[nProps].ulPropTag = PR_SENDER_ENTRYID;
			msgProps[nProps].Value.bin.cb = cbSenderEntryID;
			msgProps[nProps++].Value.bin.lpb = (BYTE*)lpSenderEntryID;
		}

		hr = lpMessage->SetProps(nProps, msgProps, NULL);
		if (hr != hrSuccess)
			goto exit;

		//Conversation topic
		if(VMHeader->hasField("Thread-Topic"))
		{
			// Thread-Topic is created as a vmime::defaultField, and is just a normal string, ignoring the text
			// so we convert the defaultField to a textField by parsing it again .. bit of a long road, but it seems to work
			vmime::defaultField *defTT = dynamic_cast <vmime::defaultField*>(VMHeader->findField("Thread-Topic"));
			vmime::textField *txtTT = dynamic_cast<vmime::textField*>(vmime::headerFieldFactory::registerer<vmime::textField>::creator());
			txtTT->parse(defTT->getValue());
			std::string convTT = txtTT->getValue().getConvertedText(MAPI_CHARSET);
			delete txtTT;

			sConTopic.Value.lpszA = (char*)convTT.c_str();
			sConTopic.ulPropTag = PR_CONVERSATION_TOPIC;

			hr = lpMessage->SetProps(1, &sConTopic, NULL);
			if (hr != hrSuccess)
				goto exit;

		} else if(HrGetOneProp(lpMessage, PR_NORMALIZED_SUBJECT, &lpPropNormalizedSubject) == hrSuccess) {
			
			sConTopic.ulPropTag = PR_CONVERSATION_TOPIC;
			sConTopic.Value.lpszA = lpPropNormalizedSubject->Value.lpszA;
			
			hr = lpMessage->SetProps(1, &sConTopic, NULL);
			if (hr != hrSuccess)
				goto exit;
		}

		// Thread-Index header
		if (VMHeader->hasField("Thread-Index"))
		{
			vmime::string outString;
			SPropValue sThreadIndex;

			string threadIndex = VMHeader->findField("Thread-Index")->getValue().generate();

			vmime::encoder* enc = vmime::encoderFactory::getInstance()->create("base64");

			vmime::utility::inputStreamStringAdapter in(threadIndex);			
			vmime::utility::outputStreamStringAdapter out(outString);

			enc->decode(in, out);

			sThreadIndex.ulPropTag = PR_CONVERSATION_INDEX;
			sThreadIndex.Value.bin.cb = outString.size();
			sThreadIndex.Value.bin.lpb = (LPBYTE)outString.c_str();

			hr = lpMessage->SetProps(1, &sThreadIndex, NULL);
			if (hr != hrSuccess)
				goto exit;
		}

		// X-Priority header
		if (VMHeader->hasField("X-Priority")) {
			SPropValue sPriority[2];
			sPriority[0].ulPropTag = PR_PRIORITY;
			sPriority[1].ulPropTag = PR_IMPORTANCE;
			string xprio = VMHeader->findField("X-Priority")->getValue().generate();
			switch (xprio[0]) {
			case '1':
			case '2':
				sPriority[0].Value.ul = PRIO_URGENT;
				sPriority[1].Value.ul = IMPORTANCE_HIGH;
				break;
			case '4':
			case '5':
				sPriority[0].Value.ul = PRIO_NONURGENT;
				sPriority[1].Value.ul = IMPORTANCE_LOW;
				break;
			default:
			case '3':
				sPriority[0].Value.ul = PRIO_NORMAL;
				sPriority[1].Value.ul = IMPORTANCE_NORMAL;
				break;
			};
			hr = lpMessage->SetProps(2, sPriority, NULL);
			if (hr != hrSuccess)
				goto exit;
		}

		// Sensitivity header
		if (VMHeader->hasField("Sensitivity")) {
			SPropValue sSensitivity[1];
			sSensitivity[0].ulPropTag = PR_SENSITIVITY;
			string sensitivity = VMHeader->findField("Sensitivity")->getValue().generate();
			if (sensitivity == "Personal") {
				sSensitivity[0].Value.ul = SENSITIVITY_PERSONAL;
			} else if (sensitivity == "Private") {
				sSensitivity[0].Value.ul = SENSITIVITY_PRIVATE;
			} else if (sensitivity == "Company-Confidential") {
				sSensitivity[0].Value.ul = SENSITIVITY_COMPANY_CONFIDENTIAL;
			} else {
				sSensitivity[0].Value.ul = SENSITIVITY_NONE;
			}
			hr = lpMessage->SetProps(1, sSensitivity, NULL);
			if (hr != hrSuccess)
				goto exit;
		}

		// read receipt	request
		if(VMHeader->hasField("Disposition-Notification-To")) {
			vmime::mailbox *mbReadReceipt = VMHeader->DispositionNotificationTo().getValue().getMailboxAt(0); // we only use the 1st
			if (mbReadReceipt && !mbReadReceipt->isEmpty()) {

				//FIXME: For zarafa use a addressbook entry?
				hr = ECCreateOneOff((LPSTR)mbReadReceipt->getName().getConvertedText(MAPI_CHARSET).c_str(), "SMTP", (LPSTR)mbReadReceipt->getEmail().c_str(), 0, &cbEntryID, &lpEntryID);
				if (hr != hrSuccess)
					goto exit;

				SPropValue sRRProps[4];
				sRRProps[0].ulPropTag = PR_READ_RECEIPT_REQUESTED;
				sRRProps[0].Value.b = true;
				
				sRRProps[1].ulPropTag = PR_MESSAGE_FLAGS;
				sRRProps[1].Value.ul = MSGFLAG_UNMODIFIED | MSGFLAG_RN_PENDING | MSGFLAG_NRN_PENDING;

				sRRProps[2].ulPropTag = PR_REPORT_ENTRYID;
				sRRProps[2].Value.bin.cb = cbEntryID;
				sRRProps[2].Value.bin.lpb = (BYTE*)lpEntryID;

				sRRProps[3].ulPropTag = PR_REPORT_NAME;
				sRRProps[3].Value.lpszA = (LPSTR)mbReadReceipt->getName().getConvertedText(MAPI_CHARSET).c_str();

				hr = lpMessage->SetProps(4, sRRProps, NULL);
				if (hr != hrSuccess)
					goto exit;

				MAPIFreeBuffer(lpEntryID); lpEntryID = NULL;
			}
		}

		std::vector<vmime::headerField*> hf = VMHeader->getFieldList();
		std::vector<vmime::headerField*>::iterator hfi;
		for (hfi = hf.begin(); hfi != hf.end(); hfi++) {
			string value, name = (*hfi)->getName();
			
			if (name[0] != 'X')
				continue;

			// exclusion list?
			if (name == "X-Priority")
				continue;

			transform(name.begin(), name.end(), name.begin(), ::tolower);

			LPMAPINAMEID lpNameID = NULL;
			LPSPropTagArray lpPropTags = NULL;

			MAPIAllocateBuffer(sizeof(MAPINAMEID), (void**)&lpNameID);
			lpNameID->lpguid = (GUID*)&PS_INTERNET_HEADERS;
			lpNameID->ulKind = MNID_STRING;

			int vlen = mbstounicode(NULL, name.c_str(), 0) +1;
			MAPIAllocateMore(vlen*sizeof(WCHAR), lpNameID, (void**)&lpNameID->Kind.lpwstrName);
			mbstounicode(lpNameID->Kind.lpwstrName, name.c_str(), vlen);

			hr = lpMessage->GetIDsFromNames(1, &lpNameID, MAPI_CREATE, &lpPropTags);
			if (hr != hrSuccess) {
				hr = hrSuccess;
				goto next;
			}

			SPropValue sProp[1];
			value = (*hfi)->getValue().generate();
			sProp[0].ulPropTag = PROP_TAG(PT_STRING8, PROP_ID(lpPropTags->aulPropTag[0]));
			sProp[0].Value.lpszA = (char*)value.c_str();

			hr = lpMessage->SetProps(1, sProp, NULL);
			if (hr != hrSuccess) {
				hr = hrSuccess;	// ignore this x-header as named props then
				goto next;
			}

next:
			if (lpPropTags)
				MAPIFreeBuffer(lpPropTags);
			lpPropTags = NULL;

			if (lpNameID)
				MAPIFreeBuffer(lpNameID);
			lpNameID = NULL;
		}
	}
	catch (vmime::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception on parsing headers: "+e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (std::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on parsing headers: %s", e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (...) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception occurred on parsing headers");
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

exit:
	if (lpRecipProps)
		MAPIFreeBuffer(lpRecipProps);

	if(lpEntryList)
		MAPIFreeBuffer(lpEntryList);

	if(lpFromEntryID)
		MAPIFreeBuffer(lpFromEntryID);

	if(lpSenderEntryID)
		MAPIFreeBuffer(lpSenderEntryID);

	if(lpEntryID)
		MAPIFreeBuffer(lpEntryID);

	if(lpPropNormalizedSubject)
		MAPIFreeBuffer(lpPropNormalizedSubject);

	return hr;
}

// Sets PR_MESSAGE_TO_ME, PR_MESSAGE_CC_ME and PR_MESSAGE_RECIP_ME appropriately
HRESULT VMIMEToMAPI::handleMessageToMeProps(IMessage *lpMessage, LPADRLIST lpRecipients)  {
    HRESULT hr = hrSuccess;
    unsigned int i = 0;
    LPSPropValue lpRecipType = NULL; // non-free
    LPSPropValue lpEntryId = NULL;	// non-free
    
    bool bToMe = false;
    bool bCcMe = false;
    bool bRecipMe = false;
    
    SPropValue sProps[3];
    
    if(m_dopt.user_entryid == NULL) {
        hr = hrSuccess; // Not an error, but don't do any processing
        goto exit;
    }
    
    // Loop through all recipients of the message to find ourselves in the recipient list.
    for(i=0;i<lpRecipients->cEntries;i++) {
        lpRecipType = PpropFindProp(lpRecipients->aEntries[i].rgPropVals, lpRecipients->aEntries[i].cValues, PR_RECIPIENT_TYPE);
        lpEntryId = PpropFindProp(lpRecipients->aEntries[i].rgPropVals, lpRecipients->aEntries[i].cValues, PR_ENTRYID);
        
        if(lpRecipType == NULL)
            continue;
            
        if(lpEntryId == NULL)
            continue;

        // The user matches if the entryid of the recipient is equal to ours            
        if(lpEntryId->Value.bin.cb != m_dopt.user_entryid->cb)
            continue;
        
        if(memcmp(lpEntryId->Value.bin.lpb, m_dopt.user_entryid->lpb, lpEntryId->Value.bin.cb) != 0)
            continue;
            
        // Users match, check what type
        bRecipMe = true;
        
        if(lpRecipType->Value.ul == MAPI_TO)
            bToMe = true;
        else if(lpRecipType->Value.ul == MAPI_CC)
            bCcMe = true;
            
    }
    
    // Set the properties
    sProps[0].ulPropTag = PR_MESSAGE_RECIP_ME;
    sProps[0].Value.b = bRecipMe;
    sProps[1].ulPropTag = PR_MESSAGE_TO_ME;
    sProps[1].Value.b = bToMe;
    sProps[2].ulPropTag = PR_MESSAGE_CC_ME;
    sProps[2].Value.b = bCcMe;
    
    lpMessage->SetProps(3, sProps, NULL);
    
exit:    
    return hr;
}

HRESULT VMIMEToMAPI::handleRecipients(vmime::header* VMHeader, IMessage* lpMessage) {
	HRESULT		hr				= hrSuccess;
	LPADRLIST 	lpRecipients	= NULL;

	try {
		vmime::addressList lpVMAListRecip = VMHeader->To().getValue();
		vmime::addressList lpVMAListCopyRecip = VMHeader->Cc().getValue();
		vmime::addressList lpVMAListBlCpRecip = VMHeader->Bcc().getValue();

		int iAdresCount = lpVMAListRecip.getAddressCount() + lpVMAListCopyRecip.getAddressCount() + lpVMAListBlCpRecip.getAddressCount();

		if (iAdresCount == 0)
			goto exit;

		hr = MAPIAllocateBuffer(CbNewADRLIST(iAdresCount), (void **)&lpRecipients);
		if (hr != hrSuccess)
			goto exit;

		lpRecipients->cEntries = 0;

		if (!lpVMAListRecip.isEmpty()) {
			hr = modifyRecipientList(lpRecipients, &lpVMAListRecip, MAPI_TO);
			if (hr != hrSuccess)
				goto exit;
		}

		if (!lpVMAListCopyRecip.isEmpty()) {
			hr = modifyRecipientList(lpRecipients, &lpVMAListCopyRecip, MAPI_CC);
			if (hr != hrSuccess)
				goto exit;
		}

		if (!lpVMAListBlCpRecip.isEmpty()) {
			hr = modifyRecipientList(lpRecipients, &lpVMAListBlCpRecip, MAPI_BCC);
			if (hr != hrSuccess)
				goto exit;
		}
		
		// Handle PR_MESSAGE_*_ME props
		hr = handleMessageToMeProps(lpMessage, lpRecipients);
		if (hr != hrSuccess)
		    goto exit;

		// actually modify recipients in mapi object
		hr = lpMessage->ModifyRecipients(MODRECIP_ADD, lpRecipients);	
		if (hr != hrSuccess)
			goto exit;

	}
	catch (vmime::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception on recipients: "+e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (std::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on recipients: %s", e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (...) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception occurred on recipients");
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

exit:
	if (lpRecipients != NULL) {
		// Because the recipient list can be altered to delete 1 row, or add 1 row using ModifyRecipients()
		// we need to free this data with FreeProws, and not the default MAPIFreeBuffer() call.
		FreeProws((LPSRowSet) lpRecipients);
	}

	return hr;
}


// Modifies a recipientlist, starting at iOffset in that list
HRESULT VMIMEToMAPI::modifyRecipientList(LPADRLIST lpRecipients, const vmime::addressList* lpVMAddressList, ULONG ulRecipType) {
	HRESULT			hr				= hrSuccess;
	int				iAddressCount	= lpVMAddressList->getAddressCount();
	ULONG			cbEntryID		= 0;
	LPENTRYID		lpEntryID		= NULL;
	vmime::mailbox	*mbx			= NULL;
	const vmime::address* lpVMAddress;
	std::string	strName, strEmail, strSearch;
	unsigned int 	iRecipNum		= 0;

	// order is important for modifyFromAddressBook()
	SizedSPropTagArray(7, sptaRecipientProps) = { 7, { PR_ADDRTYPE, PR_DISPLAY_NAME, PR_DISPLAY_TYPE, PR_EMAIL_ADDRESS, PR_ENTRYID,
													   PR_SEARCH_KEY, PR_SMTP_ADDRESS } };

	// walk through all recipients
	for (int iRecip=0; iRecip < iAddressCount; iRecip++) {
		
		try {
			mbx = NULL;
			lpVMAddress = lpVMAddressList->getAddressAt(iRecip);
			
			// not sure what to do if it IS a mailbox group.... ( never saw one )
			if (!lpVMAddress->isGroup())
				mbx = (vmime::mailbox *)lpVMAddress;

			if (!mbx)
				goto exit;

   			strEmail = mbx->getEmail();
   			// Do we have a name ? 
   			const vmime::text vmText = mbx->getName();
   			if (!vmText.isEmpty()) {
   				strName = vmText.getConvertedText(MAPI_CHARSET);
   			} else {
   				strName = strEmail;
   			}
		}
		catch (vmime::exception& e) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception on modify recipient: "+e.what());
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}
		catch (std::exception& e) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on modify recipient: %s", e.what());
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}
		catch (...) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception occurred on modify recipient");
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}

		iRecipNum = lpRecipients->cEntries;

		// use strEmail to find GAB entry
		hr = modifyFromAddressBook(&lpRecipients->aEntries[iRecipNum].rgPropVals, &lpRecipients->aEntries[iRecipNum].cValues, strEmail.c_str(), ulRecipType, (LPSPropTagArray)&sptaRecipientProps);
		if (hr != hrSuccess) {
			// Fallback if the entry was not found (or errored) in the addressbook
			int iNumTags = 8;
	
			iRecipNum = lpRecipients->cEntries;
	
			// will be cleaned up by caller.
			hr = MAPIAllocateBuffer(sizeof(SPropValue) * iNumTags, (void **)&lpRecipients->aEntries[iRecipNum].rgPropVals);	
			if (hr != hrSuccess)
				goto exit;
			
			lpRecipients->aEntries[iRecipNum].cValues						= iNumTags;	
			lpRecipients->aEntries[iRecipNum].ulReserved1					= 0;
			
			lpRecipients->aEntries[iRecipNum].rgPropVals[0].ulPropTag		= PR_RECIPIENT_TYPE;
			lpRecipients->aEntries[iRecipNum].rgPropVals[0].Value.l			= ulRecipType;
		
			lpRecipients->aEntries[iRecipNum].rgPropVals[1].ulPropTag		= PR_DISPLAY_NAME;
			hr = MAPIAllocateMore(strName.size()+1, lpRecipients->aEntries[iRecipNum].rgPropVals,
								  (void **)&lpRecipients->aEntries[iRecipNum].rgPropVals[1].Value.lpszA);
			if (hr != hrSuccess)
				goto exit;
			lstrcpy(lpRecipients->aEntries[iRecipNum].rgPropVals[1].Value.lpszA, strName.c_str());

			lpRecipients->aEntries[iRecipNum].rgPropVals[2].ulPropTag		= PR_SMTP_ADDRESS;
			hr = MAPIAllocateMore(strEmail.size()+1, lpRecipients->aEntries[iRecipNum].rgPropVals,
								  (void **)&lpRecipients->aEntries[iRecipNum].rgPropVals[2].Value.lpszA);
			if (hr != hrSuccess)
				goto exit;
			lstrcpy(lpRecipients->aEntries[iRecipNum].rgPropVals[2].Value.lpszA, strEmail.c_str());

			lpRecipients->aEntries[iRecipNum].rgPropVals[3].ulPropTag		= PR_ENTRYID;
			hr = ECCreateOneOff((char *)strName.c_str(), "SMTP", (char *)strEmail.c_str(), 0, &cbEntryID, &lpEntryID);
			if (hr != hrSuccess)
				goto exit;
				
			hr = MAPIAllocateMore(cbEntryID, lpRecipients->aEntries[iRecipNum].rgPropVals,
								  (void **)&lpRecipients->aEntries[iRecipNum].rgPropVals[3].Value.bin.lpb);
			if (hr != hrSuccess)
				goto exit;

			lpRecipients->aEntries[iRecipNum].rgPropVals[3].Value.bin.cb = cbEntryID;
			memcpy(lpRecipients->aEntries[iRecipNum].rgPropVals[3].Value.bin.lpb, lpEntryID, cbEntryID);

			lpRecipients->aEntries[iRecipNum].rgPropVals[4].ulPropTag	= PR_ADDRTYPE;
			lpRecipients->aEntries[iRecipNum].rgPropVals[4].Value.lpszA	= "SMTP";

			strSearch = "SMTP:"+strEmail;
			transform(strSearch.begin(), strSearch.end(), strSearch.begin(), ::toupper);
			lpRecipients->aEntries[iRecipNum].rgPropVals[5].ulPropTag	= PR_SEARCH_KEY;
			lpRecipients->aEntries[iRecipNum].rgPropVals[5].Value.bin.cb = strSearch.size() + 1; // we include the trailing 0 as MS does this also
			hr = MAPIAllocateMore(strSearch.size()+1, lpRecipients->aEntries[iRecipNum].rgPropVals,
								  (void **)&lpRecipients->aEntries[iRecipNum].rgPropVals[5].Value.bin.lpb);
			if (hr != hrSuccess)
				goto exit;
			memcpy(lpRecipients->aEntries[iRecipNum].rgPropVals[5].Value.bin.lpb, strSearch.c_str(), strSearch.size()+1);

			// Add Email address
			lpRecipients->aEntries[iRecipNum].rgPropVals[6].ulPropTag		= PR_EMAIL_ADDRESS;
			hr = MAPIAllocateMore(strEmail.size()+1, lpRecipients->aEntries[iRecipNum].rgPropVals,
								  (void **)&lpRecipients->aEntries[iRecipNum].rgPropVals[6].Value.lpszA);
			if (hr != hrSuccess)
				goto exit;
			lstrcpy(lpRecipients->aEntries[iRecipNum].rgPropVals[6].Value.lpszA, strEmail.c_str());

			// Add display type
			lpRecipients->aEntries[iRecipNum].rgPropVals[7].ulPropTag = PR_DISPLAY_TYPE;
			lpRecipients->aEntries[iRecipNum].rgPropVals[7].Value.ul = DT_MAILUSER;			

			if (lpEntryID) {
				MAPIFreeBuffer(lpEntryID);
				lpEntryID = NULL;
			}

		}

		lpRecipients->cEntries++;
	} // end recipient loop

exit:
	if (lpEntryID)
		MAPIFreeBuffer(lpEntryID);

	return hr;
}

// copies data from addressbook into lpRecipient
HRESULT VMIMEToMAPI::modifyFromAddressBook(LPSPropValue *lppPropVals, ULONG *lpulValues, const char *email, ULONG ulRecipType, LPSPropTagArray lpPropsList) {
	HRESULT hr = hrSuccess;
	LPENTRYID lpDDEntryID = NULL;
	ULONG cbDDEntryID;
	LPABCONT lpDefaultDir = NULL;
	ULONG ulObj = 0;
	ADRLIST *lpAdrList = NULL;
	FlagList *lpFlagList = NULL;
	LPSPropValue lpProp = NULL;
	SPropValue sRecipProps[8]; // 7 from addressbook + PR_RECIPIENT_TYPE == max
	ULONG cValues = 0;
	SizedSPropTagArray(7, sptaAddress) = { 7, { PR_SMTP_ADDRESS, PR_ADDRTYPE, PR_EMAIL_ADDRESS, PR_DISPLAY_TYPE, PR_DISPLAY_NAME, PR_ENTRYID, PR_SEARCH_KEY } };

	if (!m_lpAdrBook) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

	hr = m_lpAdrBook->GetDefaultDir(&cbDDEntryID, &lpDDEntryID);
	if (hr != hrSuccess)
		goto exit;

	hr = m_lpAdrBook->OpenEntry(cbDDEntryID, lpDDEntryID, NULL, 0, &ulObj, (LPUNKNOWN*)&lpDefaultDir);
	if (hr != hrSuccess)
		goto exit;

    hr = MAPIAllocateBuffer(CbNewSRowSet(1), (void **) &lpAdrList);
    if (hr != hrSuccess)
        goto exit;
        
    lpAdrList->cEntries = 1;
    lpAdrList->aEntries[0].cValues = 1;
    
    hr = MAPIAllocateBuffer(sizeof(SPropValue), (void **) &lpAdrList->aEntries[0].rgPropVals);
    if (hr != hrSuccess)
        goto exit;
        
    lpAdrList->aEntries[0].rgPropVals[0].ulPropTag = PR_DISPLAY_NAME;
    lpAdrList->aEntries[0].rgPropVals[0].Value.lpszA = (char *)email; // static reference is OK here

    hr = MAPIAllocateBuffer(CbNewFlagList(1), (void **) &lpFlagList);
    if (hr != hrSuccess)
        goto exit;
        
    lpFlagList->cFlags = 1;
    lpFlagList->ulFlag[0] = MAPI_UNRESOLVED;
        
    hr = lpDefaultDir->ResolveNames((LPSPropTagArray)&sptaAddress, EMS_AB_ADDRESS_LOOKUP, lpAdrList, lpFlagList);
    if (hr != hrSuccess)
        goto exit;
        
    if (lpFlagList->cFlags != 1 || lpFlagList->ulFlag[0] != MAPI_RESOLVED) {
        hr = MAPI_E_NOT_FOUND;
        goto exit;
    }

	if (PROP_TYPE(lpPropsList->aulPropTag[0]) != PT_NULL) {
		lpProp = PpropFindProp(lpAdrList->aEntries[0].rgPropVals, lpAdrList->aEntries[0].cValues, PR_ADDRTYPE);
		if (!lpProp) {
			hr = MAPI_E_NOT_FOUND;
			goto exit;
		}
		sRecipProps[cValues].ulPropTag = lpPropsList->aulPropTag[0]; // PR_xxx_ADDRTYPE;
		sRecipProps[cValues].Value.lpszA = lpProp->Value.lpszA;
		cValues++;
	}

	if (PROP_TYPE(lpPropsList->aulPropTag[1]) != PT_NULL) {
		lpProp = PpropFindProp(lpAdrList->aEntries[0].rgPropVals, lpAdrList->aEntries[0].cValues, PR_DISPLAY_NAME);
		if (!lpProp) {
			hr = MAPI_E_NOT_FOUND;
			goto exit;
		}
		sRecipProps[cValues].ulPropTag = lpPropsList->aulPropTag[1]; // PR_xxx_DISPLAY_NAME;
		sRecipProps[cValues].Value.lpszA = lpProp->Value.lpszA;
		cValues++;
	}

	if (PROP_TYPE(lpPropsList->aulPropTag[2]) != PT_NULL) {
		lpProp = PpropFindProp(lpAdrList->aEntries[0].rgPropVals, lpAdrList->aEntries[0].cValues, PR_DISPLAY_TYPE);
		if (!lpProp) {
			hr = MAPI_E_NOT_FOUND;
			goto exit;
		}
		sRecipProps[cValues].ulPropTag = lpPropsList->aulPropTag[2]; // PR_xxx_DISPLAY_TYPE;
		sRecipProps[cValues].Value.ul = lpProp->Value.ul;
		cValues++;
	}

	if (PROP_TYPE(lpPropsList->aulPropTag[3]) != PT_NULL) {
		lpProp = PpropFindProp(lpAdrList->aEntries[0].rgPropVals, lpAdrList->aEntries[0].cValues, PR_EMAIL_ADDRESS);
		if (!lpProp) {
			hr = MAPI_E_NOT_FOUND;
			goto exit;
		}
		sRecipProps[cValues].ulPropTag = lpPropsList->aulPropTag[3]; // PR_xxx_EMAIL_ADDRESS;
		sRecipProps[cValues].Value.lpszA = lpProp->Value.lpszA;
		cValues++;
	}

	if (PROP_TYPE(lpPropsList->aulPropTag[4]) != PT_NULL) {
		lpProp = PpropFindProp(lpAdrList->aEntries[0].rgPropVals, lpAdrList->aEntries[0].cValues, PR_ENTRYID);
		if (!lpProp) {
			hr = MAPI_E_NOT_FOUND;
			goto exit;
		}
		sRecipProps[cValues].ulPropTag = lpPropsList->aulPropTag[4]; // PR_xxx_ENTRYID;
		sRecipProps[cValues].Value.bin = lpProp->Value.bin;
		cValues++;
	}

	if (PROP_TYPE(lpPropsList->aulPropTag[5]) != PT_NULL) {
		lpProp = PpropFindProp(lpAdrList->aEntries[0].rgPropVals, lpAdrList->aEntries[0].cValues, PR_SEARCH_KEY);
		if (!lpProp) {
			hr = MAPI_E_NOT_FOUND;
			goto exit;
		}
		sRecipProps[cValues].ulPropTag = lpPropsList->aulPropTag[5]; // PR_xxx_SEARCH_KEY;
		sRecipProps[cValues].Value.bin = lpProp->Value.bin;
		cValues++;
	}


	if (PROP_TYPE(lpPropsList->aulPropTag[6]) != PT_NULL) {
		lpProp = PpropFindProp(lpAdrList->aEntries[0].rgPropVals, lpAdrList->aEntries[0].cValues, PR_SMTP_ADDRESS);
		if (!lpProp) {
			hr = MAPI_E_NOT_FOUND;
			goto exit;
		}
		sRecipProps[cValues].ulPropTag = lpPropsList->aulPropTag[6]; // PR_xxx_SMTP_ADDRESS;
		sRecipProps[cValues].Value.lpszA = lpProp->Value.lpszA;
		cValues++;
	}

	if (ulRecipType != MAPI_ORIG) {
		sRecipProps[cValues].ulPropTag = PR_RECIPIENT_TYPE;
		sRecipProps[cValues].Value.ul = ulRecipType;
		cValues++;
	}

	hr = Util::HrCopyPropertyArray(sRecipProps, cValues, lppPropVals, &cValues);
	if (hr == hrSuccess && lpulValues)
		*lpulValues = cValues;

	
exit:
	if (lpAdrList)
		FreeProws((LPSRowSet)lpAdrList);

	if (lpFlagList)
	    MAPIFreeBuffer(lpFlagList);

	if (lpDefaultDir)
		lpDefaultDir->Release();

	if (lpDDEntryID)
		MAPIFreeBuffer(lpDDEntryID);

	return hr;
}


/*		Disect Body

Here we are going to split the body into pieces and throw every part into its container.
We make decisions on the basis of Content Types...

Content Types...
http://www.faqs.org/rfcs/rfc2046.html

Top level				Subtypes

discrete:
	text				plain, html, richtext, enriched
	image				jpeg, gif, png.. etc
	audio				basic, wav, ai.. etc
	video				mpeg, avi.. etc
	application			octet-stream, postscript

composite:
	multipart			mixed, alternative, digest ( contains message ), paralell, 
	message				rfc822, partial ( please no fragmentation and reassembly ), external-body

*/

HRESULT VMIMEToMAPI::disectBody(vmime::header* VMHeader, vmime::body* VMBody, IMessage* lpMessage, bool onlyBody) {
	HRESULT	hr = hrSuccess;
	IStream *lpStream = NULL;

	try {
		vmime::mediaType mt = VMHeader->ContentType().getValue();

		// find body type
		if (mt.getType() == "multipart") {
			// recursively process multipart message
			// TODO: do we need to test VMBody->getPartCount() > 0 ??
			for (int i=0; i < VMBody->getPartCount(); i++) {
				vmime::bodyPart* VMBodyPart = VMBody->getPartAt(i);

				hr = disectBody(VMBodyPart->getHeader(), VMBodyPart->getBody(), lpMessage);
				if (hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to parse sub multipart %d of mail body", i);
					goto exit;
				}
			}

		// Only handle as inline text if no filename is specified and not specified as 'attachment'
		// or if the text part is the only body part in the mail
		} else if (	mt.getType() == "text" && (mt.getSubType() == "plain" || mt.getSubType() == "html") && (
					(VMHeader->ContentDisposition().getValue().getName() != "attachment" && 
					 !VMHeader->ContentDisposition().hasParameter("filename") && 
					 !VMHeader->ContentType().hasParameter("name")) || 
					 onlyBody)
					) {
			if (mt.getSubType() == "html") {
				hr = handleHTMLTextpart(VMHeader, VMBody, lpMessage);
				if (hr != hrSuccess) {
					lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to parse mail HTML text");
					goto exit;
				}
			} else {
				hr = handleTextpart(VMHeader, VMBody, lpMessage);
				if (hr != hrSuccess)
					goto exit;
			}
		
		} else if (mt.getType() == "message") {
			// Create Attach			
			ULONG		ulAttNr			= 0;
			LPATTACH	pAtt			= NULL;
			IMessage*	lpNewMessage	= NULL;
			LPSPropValue	lpSubject	= NULL;
			char *		lpszBody	= NULL;
			char *		lpszBodyOrig	= NULL;

			std::string newMessage;
			vmime::utility::outputStreamStringAdapter os(newMessage);
			VMBody->generate(os);
					
			lpszBodyOrig = lpszBody = (char *)newMessage.c_str();

			// Skip any leading newlines from the e-mail (attached messaged produced by Microsoft MimeOLE seem to do this)
			while (*lpszBody != '\0' && (*lpszBody == '\r' || *lpszBody == '\n'))
				lpszBody++;

			// and remove from string
			newMessage.erase(0, lpszBody - lpszBodyOrig);

			hr = lpMessage->CreateAttach(0, 0, &ulAttNr, &pAtt);
			if (hr != hrSuccess)
				goto next;

			hr = pAtt->OpenProperty(PR_ATTACH_DATA_OBJ, &IID_IMessage, 0, MAPI_CREATE | MAPI_MODIFY, (LPUNKNOWN *)&lpNewMessage);
			if (hr != hrSuccess)
				goto next;

			hr = convertVMIMEToMAPI(newMessage, lpNewMessage);
			if (hr != hrSuccess)
				goto next;

			hr = HrGetOneProp(lpNewMessage, PR_SUBJECT, &lpSubject);
			if (hr != hrSuccess)
				goto next;

			// Set PR_ATTACH_FILENAME of attachment to message subject, (WARNING: abuse of lpSubject variable)
			lpSubject->ulPropTag = PR_DISPLAY_NAME;
			pAtt->SetProps(1, lpSubject, NULL);

			lpSubject->ulPropTag = PR_ATTACH_METHOD;
			lpSubject->Value.ul = ATTACH_EMBEDDED_MSG;
			pAtt->SetProps(1, lpSubject, NULL);

			lpNewMessage->SaveChanges(0);
			pAtt->SaveChanges(0);

next:
			if (lpSubject) {
				MAPIFreeBuffer(lpSubject);
				lpSubject = NULL;
			}

			if (lpNewMessage) {
				lpNewMessage->Release();
				lpNewMessage = NULL;
			}

			if (pAtt) {
				pAtt->Release();
				pAtt = NULL;
			}
			
			if (lpStream) {
				lpStream->Release();
				lpStream = NULL;
			}

		} else if(mt.getType() == "application" && mt.getSubType() == "ms-tnef") {
			LARGE_INTEGER zero = {{0,0}};
			
			hr = CreateStreamOnHGlobal(NULL, TRUE, &lpStream);	
			if(hr != hrSuccess)
				goto exit;
				
			outputStreamMAPIAdapter str(lpStream);
			VMBody->getContents().extract(str);
			hr = lpStream->Seek(zero, STREAM_SEEK_SET, NULL);
			if(hr != hrSuccess)
				goto exit;
			
			ECTNEF tnef(TNEF_DECODE, lpMessage, lpStream);

			hr = tnef.ExtractProps(TNEF_PROP_EXCLUDE, NULL);
			if (hr == hrSuccess) {
				tnef.Finish();
			} else {
				lpLogger->Log(EC_LOGLEVEL_WARNING, "TNEF attachment parsing failed: 0x%08X", hr);
				hr = hrSuccess;
			}
		} else if (PARSE_ICAL && mt.getType() == "text" && mt.getSubType() == "calendar") {
			// ical file
			string icaldata;
			vmime::utility::outputStreamStringAdapter os(icaldata);
			VMBody->getContents().extract(os);

			hr = HrICalendar2MAPI(icaldata.c_str(), lpMessage, VMBody->getCharset().getName());
			if (hr != hrSuccess) {
				lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to parse ical information: 0x%08X, adding as normal attachment", hr);

				hr = handleAttachment(VMHeader, VMBody, lpMessage);
				if (hr != hrSuccess)
					goto exit;
			}
		} else if (mt.getType() == "application" && mt.getSubType() == "applefile") {
		} else if (mt.getType() == "application" && mt.getSubType() == "mac-binhex40") {
				// ignore appledouble parts
				// mac-binhex40 is appledouble v1, applefile is v2
				// see: http://www.iana.org/assignments/media-types/multipart/appledouble
		} else if (mt.getType() == "application" && mt.getSubType() == "octet-stream") {
			if (VMHeader->ContentDisposition().hasParameter("filename") || VMHeader->ContentType().hasParameter("name")) {
				// should be attachment
				hr = handleAttachment(VMHeader, VMBody, lpMessage);
				if (hr != hrSuccess)
					goto exit;
			} else {
				// possibly text?
				hr = handleTextpart(VMHeader, VMBody, lpMessage);
				if (hr != hrSuccess)
					goto exit;
			}
		} else {
			hr = handleAttachment(VMHeader, VMBody, lpMessage);
			if (hr != hrSuccess)
				goto exit;
		}
	}
	catch (vmime::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception on parsing body: "+e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (std::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on parsing body: %s", e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (...) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception occurred on parsing body");
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

exit:
	if(lpStream)
		lpStream->Release();
		
	return hr;
}


HRESULT VMIMEToMAPI::handleTextpart(vmime::header* VMHeader, vmime::body* VMBody, IMessage* lpMessage) {
	HRESULT hr = S_OK;
	std::string strBuffOut;

	if (bodyLevel < BODY_PLAIN) {
		// we have no body up until now, not even an HTML body
		try {
			std::string strMAPIText;
			SPropValue body[1];
			
			vmime::utility::outputStreamStringAdapter os(strBuffOut);
			VMBody->getContents().extract(os);
						
			// Convert the body to something MAPI is expecting to have in PR_BODY
			try {
				vmime::charset bodyCharset;

				// iso-8859-15 is compatible with us-ascii. We do this
				// so that some broken e-mail with no charset information
				// still come through with accented characters etc.
				bodyCharset = VMBody->getCharset();
				if(bodyCharset.getName() == vmime::charsets::US_ASCII)
					bodyCharset = vmime::charset(vmime::charsets::ISO8859_15);

				vmime::charset::convert(strBuffOut, strMAPIText, bodyCharset, MAPI_CHARSET);
			}
			catch (vmime::exception&) {
				// vmime::charset::convert() threw up, paste 'raw' body in PR_BODY to deliver the e-mail
				strMAPIText = strBuffOut;
			}

			body[0].ulPropTag = PR_BODY;
			body[0].Value.lpszA = (char*)strMAPIText.c_str();
			hr = lpMessage->SetProps(1, body, NULL);
			if (hr != hrSuccess)
				goto exit;			
		}
		catch (vmime::exception& e) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception on text body: "+e.what());
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}
		catch (std::exception& e) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on text body: %s", e.what());
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}
		catch (...) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception occurred on text body");
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}
		bodyLevel = BODY_PLAIN;
	} else {
		// we already had a plaintext or html body, so attach this text part
		hr = handleAttachment(VMHeader, VMBody, lpMessage);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to parse attached text mail");
			goto exit;
		}
	}

exit:

	return hr;
}

// put textpart into pr_rtf_compressed
HRESULT VMIMEToMAPI::handleHTMLTextpart(vmime::header* VMHeader, vmime::body* VMBody, IMessage* lpMessage) {
	std::string strRtfOut, out;
	HRESULT		hr				= hrSuccess;
	IStream*	lpHTMLStream	= NULL; 
	ULONG		cbWritten		= 0;
	std::string strHTML;

	if (bodyLevel < BODY_HTML) {
		// we're overriding a plain text body, or setting a new body
		try {
			vmime::charset bodyCharset;
			std::basic_string<WCHAR> wstrHTML;

			vmime::utility::outputStreamStringAdapter os(strHTML);
			VMBody->getContents().extract(os);
		
			// iso-8859-15 is compatible with us-ascii. We do this
			// so that some broken e-mail with no charset information
			// still come through with accented characters etc.
			bodyCharset = VMBody->getCharset();
			if(bodyCharset.getName() == vmime::charsets::US_ASCII)
				bodyCharset = vmime::charset(vmime::charsets::ISO8859_15);
		
			// convert incoming HTML source into WCHARs
			if((hr = HrConvertToWString(strHTML, bodyCharset.getName(), wstrHTML)) != hrSuccess) {
				hr = HrConvertToWString(strHTML, vmime::charsets::ISO8859_15, wstrHTML);
			}
			
			if(hr != hrSuccess)
				goto exit;

			// convert wchar-HTML into single-byte us-ascii HTML with escaped chars
			hr = HrEscapeHTML(wstrHTML, strHTML);
			if(hr != hrSuccess)
				goto exit;
		}
		catch (vmime::exception& e) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception on html body: "+e.what());
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}
		catch (std::exception& e) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on html body: %s", e.what());
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}
		catch (...) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception occurred on html body");
			hr = MAPI_E_CALL_FAILED;
			goto exit;
		}

		hr = lpMessage->OpenProperty(PR_HTML, &IID_IStream, STGM_TRANSACTED, MAPI_MODIFY|MAPI_CREATE,(LPUNKNOWN *)&lpHTMLStream);
		if (hr != hrSuccess)
			goto exit;

		// reset old body if we just set the plain text version
		ULARGE_INTEGER l = {{0,0}};
		hr = lpHTMLStream->SetSize(l);
		if (hr != hrSuccess)
			goto exit;

		hr = lpHTMLStream->Write(strHTML.c_str(), strHTML.length(), &cbWritten);
		if (hr != hrSuccess)		// check cbWritten too?
			goto exit;

		bodyLevel = BODY_HTML;
	} else {
		// already found html as body, so this is an attachment
		hr = handleAttachment(VMHeader, VMBody, lpMessage);
		if (hr != hrSuccess) {
			lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to parse attached text mail");
			goto exit;
		}
	}

exit:
	if( lpHTMLStream ){
		hr = lpHTMLStream->Commit(0);
		lpHTMLStream->Release(); 
	}

	return hr;
}


// Handle Attachments.. Now works for inlines and attachments...
HRESULT VMIMEToMAPI::handleAttachment(vmime::header* VMHeader, vmime::body* VMBody, IMessage* lpMessage) {
	HRESULT		hr			= hrSuccess;
	IStream		*lpStream	= NULL; 
	LPATTACH	lpAtt		= NULL;
	ULONG		ulAttNr		= 0;
	std::string	strLongFilename, strFilename, strId, strMimeType;
	int			nProps = 0;
	SPropValue	attProps[10];	// only 7 used currently
	int reparse = 0;
	

	// Create Attach
	hr = lpMessage->CreateAttach(0, 0, &ulAttNr, &lpAtt);
	if (hr != hrSuccess)
		goto exit;

	// open stream
	hr = lpAtt->OpenProperty(PR_ATTACH_DATA_BIN, &IID_IStream, STGM_WRITE|STGM_TRANSACTED,
							MAPI_CREATE|MAPI_MODIFY, (LPUNKNOWN *)&lpStream);
	if (hr != hrSuccess)
		goto exit;

	try {
		// attach adapter, generate in right encoding
		outputStreamMAPIAdapter osMAPI(lpStream);
		VMBody->getContents().generate(osMAPI, vmime::encoding(vmime::encodingTypes::BINARY));

		hr = lpStream->Commit(0);
		if (hr != hrSuccess)
			goto exit;
			
        // Free memory used by the stream
        lpStream->Release();
        lpStream = NULL;

		// set info on attachment
		attProps[nProps].ulPropTag = PR_ATTACH_METHOD;
		attProps[nProps++].Value.ul = ATTACH_BY_VALUE;

		// filenames
		if (VMHeader->ContentDisposition().hasParameter("filename")) {
			strLongFilename = VMHeader->ContentDisposition().getFilename().getConvertedText(MAPI_CHARSET);
			if (VMHeader->ContentDisposition().getFilename().getCharset() == vmime::charset(vmime::charsets::US_ASCII))
				reparse = 1;
		} else if (VMHeader->ContentType().hasParameter("name")) {
			strLongFilename = VMHeader->ContentType().getParameter("name")->getValue().generate();
			reparse = 1;		// never seen q-p here, but does not hurt to try
		} else {
			strLongFilename = "inline.txt";
		}

		// quoted-printable is not available in vmime 0.7.1 on parameterizedHeaderField class
		// outlook express sends qp, icedove sends the encoded version (filename*=<charset>''%FF)
		if (reparse) {
			// text was in plain text, so convert possible quoted-printable
			vmime::word wFilename;
			wFilename.parse(strLongFilename);
			strLongFilename = wFilename.getConvertedText(MAPI_CHARSET);
		}

		getShortFilename(strLongFilename, strFilename);

		attProps[nProps].ulPropTag = PR_ATTACH_LONG_FILENAME;
		attProps[nProps++].Value.lpszA = (char*)strLongFilename.c_str();
		attProps[nProps].ulPropTag = PR_ATTACH_FILENAME;
		attProps[nProps++].Value.lpszA = (char*)strFilename.c_str();

		// i get the impression this doesn't work... so do we really need to set this?
		attProps[nProps].ulPropTag = PR_RENDERING_POSITION;
		attProps[nProps++].Value.i = -1;

		strId = VMHeader->ContentId().getValue().getId();
		if (!strId.empty()) {
			// only set this property when string is present
			// otherwise, you don't get the 'save attachments' list in the main menu of outlook
			attProps[nProps].ulPropTag = PR_ATTACH_CONTENT_ID;
			attProps[nProps++].Value.lpszA = (char*)strId.c_str();
		}

		try {
			if (!VMHeader->ContentType().getValue().getType().empty() && !VMHeader->ContentType().getValue().getSubType().empty()) {
				strMimeType = VMHeader->ContentType().getValue().getType() + "/" + VMHeader->ContentType().getValue().getSubType();
				attProps[nProps].ulPropTag = PR_ATTACH_MIME_TAG;
				attProps[nProps++].Value.lpszA = (char*)strMimeType.c_str();
			}
		}
		catch (vmime::exceptions::no_such_field) {
		}

		hr = lpAtt->SetProps(nProps, attProps, NULL);
		if (hr != hrSuccess)
			goto exit;
	}
	catch (vmime::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "VMIME exception on attachment: "+e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (std::exception& e) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "STD exception on attachment: %s", e.what());
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}
	catch (...) {
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unknown generic exception occurred on attachment");
		hr = MAPI_E_CALL_FAILED;
		goto exit;
	}

	hr = lpAtt->SaveChanges(0);
	if (hr != hrSuccess)
		goto exit;

exit:
	if (hr != hrSuccess)
		lpLogger->Log(EC_LOGLEVEL_FATAL, "Unable to create attachment");

	if (lpAtt)
		lpAtt->Release();

	if (lpStream)
		lpStream->Release();

	return hr;
}

// options.h code
void imopt_default_delivery_options(delivery_options *dopt) {
	dopt->use_received_date = true;
	dopt->mark_as_read = false;
	dopt->user_entryid = NULL;
}

void imopt_default_sending_options(sending_options *sopt) {
	sopt->alternate_boundary = NULL;
	sopt->no_recipients_workaround = false;
}
