/*
 * This file is (c) Copyright Zarafa, The Netherlands, 2007
 *
 * Use of this file and the source contained within it is subject to the
 * license of the package it belongs to. See the LICENSE file in the
 * top-level directory of the package.
 *
 */

/*
 * This is our TNEF class, which has been specially designed for
 * simple TNEF reading and writing. Most importantly, we currently
 * only support properties within the message itself, and
 * do not support recipient tables and attachment properties within
 * the TNEF data. 
 *
 * The ECTNEF class will accept TNEF streams with data about
 * recipients and attachments, but the information will be ignored.
 *
 * TNEF has gone through various versions for Microsoft Mail and
 * other really old systems, and therefore has an elaborate backwards-
 * compatibility system. This means that most properties can be stored
 * in both TNEF properties as within a single 'MAPI' property of the
 * TNEF stream (0x00069003). We basically discard all the backwards-
 * compatibility and write TNEF streams that only work with
 * Outlook 2000 or later (maybe also Outlook 97, not sure about that)
 * by only writing the TNEF stream properties in 0x00069003.
 *
 * -- Steve
 */
 
#include "platform.h"

#include <mapidefs.h> 
#include <mapiutil.h>
#include "tnef.h"

// The mapping between Microsoft Mail IPM classes and those used in MAPI
// see: http://msdn2.microsoft.com/en-us/library/ms527360.aspx
struct _sClassMap {
	const char *szScheduleClass;
	const char *szMAPIClass;
} sClassMap[] = {
	{ "IPM.Microsoft Schedule.MtgReq",		"IPM.Schedule.Meeting.Request" },
	{ "IPM.Microsoft Schedule.MtgRespP",	"IPM.Schedule.Meeting.Resp.Pos" },
	{ "IPM.Microsoft Schedule.MtgRespN",	"IPM.Schedule.Meeting.Resp.Neg" },
	{ "IPM.Microsoft Schedule.MtgRespA",	"IPM.Schedule.Meeting.Resp.Tent" },
	{ "IPM.Microsoft Schedule.MtgCncl",		"IPM.Schedule.Meeting.Canceled" },
	{ "IPM.Microsoft Mail.Non-Delivery",	"Report.IPM.Note.NDR" },
	{ "IPM.Microsoft Mail.Read Receipt",	"Report.IPM.Note.IPNRN" },
	{ "IPM.Microsoft Mail.Note",			"IPM.Note" },
	{ "IPM.Microsoft Mail.Note",			"IPM" }
};

const char *FindMAPIClassByScheduleClass(char *szSClass) {
	unsigned int i = 0;

	for(i=0;i<sizeof(sClassMap)/sizeof(sClassMap[0]);i++) {
		if(stricmp(szSClass, sClassMap[i].szScheduleClass) == 0) {
			return sClassMap[i].szMAPIClass;
		}
	}

	return NULL;
}

const char *FindScheduleClassByMAPIClass(char *szMAPIClass) {
	unsigned int i = 0;

	for(i=0;i<sizeof(sClassMap)/sizeof(sClassMap[0]);i++) {
		if(stricmp(szMAPIClass, sClassMap[i].szMAPIClass) == 0) {
			return sClassMap[i].szScheduleClass;
		}
	}

	return NULL;
}

// Returns TRUE if the given property tag is in the given property tag array
bool PropTagInPropList(ULONG ulPropTag, LPSPropTagArray lpPropList)
{
	unsigned int i;

	if(lpPropList == NULL)
		return false;

	for(i=0;i<lpPropList->cValues;i++) {
		if(ulPropTag == lpPropList->aulPropTag[i]) 
			return true;
	}

	return false;
}

ECTNEF::ECTNEF(ULONG ulFlags, IMessage *lpMessage, IStream *lpStream)
{
	this->ulFlags = ulFlags;
	this->lpMessage = lpMessage;
	this->lpStream = lpStream;
}

ECTNEF::~ECTNEF()
{
	std::list<SPropValue *>::iterator iterProps;
	std::list<tnefattachment*>::iterator iterAttach;

	for(iterProps = lstProps.begin(); iterProps != lstProps.end(); iterProps++) {
		MAPIFreeBuffer(*iterProps);
	}
	for (iterAttach = lstAttachments.begin(); iterAttach != lstAttachments.end(); iterAttach++) {
		delete [] (*iterAttach)->data;
		for (iterProps = (*iterAttach)->lstProps.begin(); iterProps != (*iterAttach)->lstProps.end(); iterProps++) {
			MAPIFreeBuffer(*iterProps);
		}
	}
}

HRESULT StreamToPropValue(IStream* lpStream, ULONG ulPropTag, LPSPropValue* lppPropValue)
{
	HRESULT			hr = hrSuccess;
	LPSPropValue	lpPropValue = NULL;
	STATSTG			sStatstg;
	ULONG			ulRead = 0;
	ULONG			ulReaded = 0;

	if(lpStream == NULL || lppPropValue == NULL)
	{
		hr = MAPI_E_INVALID_OBJECT;
		goto exit;
	}

	if(PROP_TYPE(ulPropTag) != PT_BINARY || PROP_TYPE(ulPropTag) != PT_STRING8)
	{
		hr = MAPI_E_INVALID_TYPE;
		goto exit;
	}

	hr = lpStream->Stat(&sStatstg, 0);
	if(hr != hrSuccess)
		goto exit;

	hr = MAPIAllocateBuffer(sizeof(LPSPropValue), (void**)&lpPropValue);
	if(hr != hrSuccess)
		goto exit;

	lpPropValue->ulPropTag = ulPropTag;
	
	if(PROP_TYPE(ulPropTag) == PT_BINARY ) {
		lpPropValue->Value.bin.cb = (ULONG)sStatstg.cbSize.QuadPart;
	
		hr = MAPIAllocateMore((ULONG)sStatstg.cbSize.QuadPart, lpPropValue, (void**)&lpPropValue->Value.bin.lpb);
		if(hr != hrSuccess)
			goto exit;
	}else if(PROP_TYPE(ulPropTag) == PT_STRING8 ) {
		hr = MAPIAllocateMore((ULONG)sStatstg.cbSize.QuadPart, lpPropValue, (void**)&lpPropValue->Value.lpszA);
		if(hr != hrSuccess)
			goto exit;
	}

	while(1) {
		if(PROP_TYPE(ulPropTag) == PT_BINARY)
			hr = lpStream->Read(lpPropValue->Value.bin.lpb + ulReaded, 4096, &ulRead);
		else if(PROP_TYPE(ulPropTag) == PT_STRING8 )
			hr = lpStream->Read(lpPropValue->Value.lpszA + ulReaded, 4096, &ulRead);

		if(hr != hrSuccess)
			goto exit;

		if(ulRead == 0) {
			break;
		}
		ulReaded += ulRead;
	}

	*lppPropValue = lpPropValue;
exit:
	if(hr != hrSuccess && lpPropValue)
		MAPIFreeBuffer(lpPropValue);

	return hr;
}

// Adds the requested properties from the message into the pending TNEF stream
HRESULT ECTNEF::AddProps(ULONG ulFlags, LPSPropTagArray lpPropList)
{
	HRESULT			hr = hrSuccess;
	LPSPropTagArray lpPropListMessage = NULL;
	LPSPropValue	lpPropValue = NULL; // do not free at exit
	SPropTagArray	sPropTagArray;
	unsigned int	i = 0;
	bool			fPropTagInList = false;
	ULONG			cValue = 0;
	IStream*		lpStream = NULL;

	// Loop through all the properties on the message, and only
	// add those that we want to add to the list

	hr = lpMessage->GetPropList(0, &lpPropListMessage);
	if(hr != hrSuccess)
		goto exit;

	for(i=0;i<lpPropListMessage->cValues;i++) {
		fPropTagInList = PropTagInPropList(lpPropListMessage->aulPropTag[i], lpPropList);

		if(PROP_TYPE(lpPropListMessage->aulPropTag[i]) == PT_OBJECT || 
		   PROP_TYPE(lpPropListMessage->aulPropTag[i]) == PT_UNSPECIFIED ||
		   PROP_TYPE(lpPropListMessage->aulPropTag[i]) == PT_NULL)
			continue;

		if(( (ulFlags & TNEF_PROP_INCLUDE) && fPropTagInList) || ((ulFlags & TNEF_PROP_EXCLUDE) && !fPropTagInList)) {

			sPropTagArray.cValues = 1;
			sPropTagArray.aulPropTag[0] = lpPropListMessage->aulPropTag[i];
			hr = lpMessage->GetProps(&sPropTagArray, 0, &cValue, &lpPropValue);

			if(hr == hrSuccess) {
				lstProps.push_back(lpPropValue);
			} if(hr == MAPI_W_ERRORS_RETURNED && lpPropValue != NULL && lpPropValue->Value.err == MAPI_E_NOT_ENOUGH_MEMORY) {
				if(lpMessage->OpenProperty(lpPropListMessage->aulPropTag[i], &IID_IStream, 0, 0, (LPUNKNOWN *)&lpStream) == hrSuccess)
				{

					hr = StreamToPropValue(lpStream, lpPropListMessage->aulPropTag[i], &lpPropValue);
					if(hr == hrSuccess)
						lstProps.push_back(lpPropValue);

					lpStream->Release(); lpStream = NULL;
				}
			}
		
			hr = hrSuccess; // silently ignore the property
		
		}
	}

exit:
	if(lpPropListMessage)
		MAPIFreeBuffer(lpPropListMessage);

	return hr;
}

// Extracts the properties from the TNEF stream, and sets them in the message
HRESULT	ECTNEF::ExtractProps(ULONG ulFlags, LPSPropTagArray lpPropList)
{
	HRESULT hr = hrSuccess;
	ULONG ulSignature = 0;
	ULONG ulType = 0;
	ULONG ulSize = 0;
	unsigned short ulChecksum = 0;
	unsigned short ulKey = 0;
	unsigned char ulComponent = 0;
	BYTE *lpBuffer = NULL;
	SPropValue sProp;
	char *szSClass = NULL;
	// Attachments props
	LPSPropValue lpProp;
	tnefattachment* lpTnefAtt = NULL;

	hr = HrReadDWord(lpStream, &ulSignature);
	if(hr != hrSuccess)
		goto exit;

	// Check signature
	if(ulSignature != TNEF_SIGNATURE) {
		hr = MAPI_E_CORRUPT_DATA;
		goto exit;
	}

	hr = HrReadWord(lpStream, &ulKey);
	if(hr != hrSuccess)
		goto exit;

	// File is made of blocks, with each a type and size. Component and Key are ignored.
	while(1) {
		hr = HrReadByte(lpStream, &ulComponent);

		if(hr != hrSuccess) {
			hr = hrSuccess; // EOF -> no error
			goto exit;
		}

		hr = HrReadDWord(lpStream, &ulType);
		if(hr != hrSuccess)
			goto exit;

		hr = HrReadDWord(lpStream, &ulSize);
		if(hr != hrSuccess)
			goto exit;

		if (ulSize == 0) {
			// do not allocate 0 size data block
			hr = MAPI_E_CORRUPT_DATA;
			goto exit;
		}

		hr = MAPIAllocateBuffer(ulSize, (void **)&lpBuffer);
		if(hr != hrSuccess)
			goto exit;

		hr = HrReadData(lpStream, (char *)lpBuffer, ulSize);
		if(hr != hrSuccess)
			goto exit;

		hr = HrReadWord(lpStream, &ulChecksum);
		if(hr != hrSuccess)
			goto exit;

		// Loop through all the blocks of the TNEF data. We are only interested
		// in the properties block for now (0x00069003)

		switch(ulType) {
		case 0x00069003: 
			hr = HrReadPropStream((char *)lpBuffer, ulSize, lstProps);
			if (hr != hrSuccess)
				goto exit;
			break;
		case 0x00078008: /* PR_MESSAGE_CLASS */
			{
				szSClass = new char[ulSize+1];
				char *szMAPIClass = NULL;

				// NULL terminate the string
				memcpy(szSClass, lpBuffer, ulSize);
				szSClass[ulSize] = 0;

				// We map the Schedule+ message class to the more modern MAPI message
				// class. The mapping should be correct as far as we can find ..

				szMAPIClass = (char *)FindMAPIClassByScheduleClass(szSClass);
				if(szMAPIClass == NULL)
					szMAPIClass = szSClass;	// mapping not found, use string from TNEF file

				sProp.ulPropTag = PR_MESSAGE_CLASS;
				sProp.Value.lpszA = szMAPIClass;

				// We do a 'SetProps' now because we want to override the PR_MESSAGE_CLASS
				// setting, while Finish() never overrides already-present properties for
				// security reasons.

				lpMessage->SetProps(1, &sProp, NULL);

				delete [] szSClass;
				szSClass = NULL;

				break;
			}
		case 0x00050008: /* PR_OWNER_APPT_ID */
			if(ulSize == 4 && lpBuffer) {
				sProp.ulPropTag = PR_OWNER_APPT_ID;
				sProp.Value.l = *((long*)lpBuffer);
				lpMessage->SetProps(1, &sProp, NULL);
			}
			break;
		case 0x00040009:/* PR_RESPONSE_REQUESTED */
			if(ulSize == 2 && lpBuffer) {
				sProp.ulPropTag = PR_RESPONSE_REQUESTED;
				sProp.Value.b = (bool)(*(short*)lpBuffer);
				lpMessage->SetProps(1, &sProp, NULL);
			}
			break;

// --- TNEF attachemnts ---
		case 0x00069002:		// Start marker of attachment
			// data not needed .. i guess .. no idea what it should represent, 14 bytes of crap
			if (lpTnefAtt && lpTnefAtt->data) {
				// end marker previous attachment
				lstAttachments.push_back(lpTnefAtt);
			}
			lpTnefAtt = new tnefattachment;
			lpTnefAtt->size = 0;
			lpTnefAtt->data = NULL;
			break;

		case 0x00018010:		// PR_ATTACH_FILENAME
			if (!lpTnefAtt) {
				hr = MAPI_E_CORRUPT_DATA;
				goto exit;
			}

			MAPIAllocateBuffer(sizeof(SPropValue), (void**)&lpProp);
			lpProp->ulPropTag = PR_ATTACH_FILENAME;
			MAPIAllocateMore(ulSize, lpProp, (void**)&lpProp->Value.lpszA);
			memcpy(lpProp->Value.lpszA, lpBuffer, ulSize);
			lpTnefAtt->lstProps.push_back(lpProp);
			break;

		case 0x00068011:		// PR_ATTACH_RENDERING, extra icon information
			if (!lpTnefAtt) {
				hr = MAPI_E_CORRUPT_DATA;
				goto exit;
			}

			MAPIAllocateBuffer(sizeof(SPropValue), (void**)&lpProp);
			lpProp->ulPropTag = PR_ATTACH_RENDERING;
			MAPIAllocateMore(ulSize, lpProp, (void**)&lpProp->Value.bin.lpb);
			lpProp->Value.bin.cb = ulSize;
			memcpy(lpProp->Value.bin.lpb, lpBuffer, ulSize);
			lpTnefAtt->lstProps.push_back(lpProp);
			break;

		case 0x0006800f:		// PR_ATTACH_DATA_BIN, will be set via OpenProperty() in ECTNEF::Finish()
			if (!lpTnefAtt) {
				hr = MAPI_E_CORRUPT_DATA;
				goto exit;
			}
			lpTnefAtt->size = ulSize;
			lpTnefAtt->data = new BYTE[ulSize];
			memcpy(lpTnefAtt->data, lpBuffer, ulSize);
			break;

		case 0x00069005:		// Attachment property stream
			if (!lpTnefAtt) {
				hr = MAPI_E_CORRUPT_DATA;
				goto exit;
			}
			hr = HrReadPropStream((char *)lpBuffer, ulSize, lpTnefAtt->lstProps);
			if (hr != hrSuccess)
				goto exit;
			break;

		default:
			// Ignore this block
			break;
		}

		MAPIFreeBuffer(lpBuffer);
		lpBuffer = NULL;
	}

exit:
	if (lpTnefAtt && lpTnefAtt->data)	// attachment should be complete before adding
		lstAttachments.push_back(lpTnefAtt);

	if(szSClass)
		delete [] szSClass;

	if(lpBuffer)
		MAPIFreeBuffer(lpBuffer);

	return hr;
}

// Write a complete property stream (simply
HRESULT ECTNEF::HrWritePropStream(IStream *lpStream)
{
	HRESULT hr = hrSuccess;
	std::list<SPropValue *>::iterator iterProps;

	hr = HrWriteDWord(lpStream, lstProps.size());
	if(hr != hrSuccess)
		goto exit;

	for (iterProps = lstProps.begin(); iterProps != lstProps.end(); iterProps++) {
		hr = HrWriteSingleProp(lpStream, *iterProps);
		if (hr != hrSuccess)
			goto exit;
	}

exit:
	return hr;
}

HRESULT ECTNEF::HrWriteSingleProp(IStream *lpStream, LPSPropValue lpProp) 
{
	HRESULT hr = hrSuccess;
	SPropTagArray sPropTagArray;
	LPSPropTagArray lpsPropTagArray = &sPropTagArray;
	ULONG cNames = 0;
	MAPINAMEID **lppNames = NULL;
	ULONG ulLen = 0;
	ULONG ulMVProp = 0;
	ULONG ulCount = 0;

	if(PROP_ID(lpProp->ulPropTag) >= 0x8000) {
		// Get named property GUID and ID or name
		sPropTagArray.cValues = 1;
		sPropTagArray.aulPropTag[0] = lpProp->ulPropTag;

		hr = lpMessage->GetNamesFromIDs(&lpsPropTagArray, NULL, 0, &cNames, &lppNames);
		if(hr != hrSuccess) {
			hr = hrSuccess; // Ignore this property
			goto exit;
		}

		if(cNames == 0 || lppNames == NULL || lppNames[0] == NULL) {
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}

		// Write the property tag
		hr = HrWriteDWord(lpStream, lpProp->ulPropTag);
		if(hr != hrSuccess)
			goto exit;

		hr = HrWriteData(lpStream, (char *)lppNames[0]->lpguid, sizeof(GUID));
		if(hr != hrSuccess)
			goto exit;

		if(lppNames[0]->ulKind == MNID_ID) {
			hr = HrWriteDWord(lpStream, 0);
			if(hr != hrSuccess)
				goto exit;

			hr = HrWriteDWord(lpStream, lppNames[0]->Kind.lID);
			if(hr != hrSuccess)
				goto exit;
		} else {
			hr = HrWriteDWord(lpStream, 1);
			if(hr != hrSuccess)
				goto exit;

			ulLen = unicodelen(lppNames[0]->Kind.lpwstrName)*sizeof(WCHAR)+sizeof(WCHAR);

			hr = HrWriteDWord(lpStream, ulLen);
			if(hr != hrSuccess)
				goto exit;

			hr = HrWriteData(lpStream, (char *)lppNames[0]->Kind.lpwstrName, ulLen);
			if(hr != hrSuccess)
				goto exit;

			// Align to 4-byte boundary
			while(ulLen & 3) {
				hr = HrWriteByte(lpStream, 0);
				if(hr != hrSuccess)
					goto exit;
				ulLen++;
			}
		}
	} else {
		// Write the property tag
		hr = HrWriteDWord(lpStream, lpProp->ulPropTag);
		if(hr != hrSuccess)
			goto exit;
	}

	// Now, write the actual property value
	if(PROP_TYPE(lpProp->ulPropTag) & MV_FLAG) {
		switch(PROP_TYPE(lpProp->ulPropTag)) {
		case PT_MV_I2:
			ulCount = lpProp->Value.MVi.cValues;
			break;
		case PT_MV_LONG:
			ulCount = lpProp->Value.MVl.cValues;
			break;
		case PT_MV_R4:
			ulCount = lpProp->Value.MVflt.cValues;
			break;
		case PT_MV_APPTIME:
			ulCount = lpProp->Value.MVat.cValues;
			break;
		case PT_MV_DOUBLE:
			ulCount = lpProp->Value.MVdbl.cValues;
			break;
		case PT_MV_CURRENCY:
			ulCount = lpProp->Value.MVcur.cValues;
			break;
		case PT_MV_SYSTIME:
			ulCount = lpProp->Value.MVft.cValues;
			break;
		case PT_MV_I8:
			ulCount = lpProp->Value.MVli.cValues;
			break;
		case PT_MV_STRING8:
			ulCount = lpProp->Value.MVszA.cValues;
			break;
		case PT_MV_UNICODE:
			ulCount = lpProp->Value.MVszW.cValues;
			break;
		case PT_MV_BINARY:
			ulCount = lpProp->Value.MVbin.cValues;
			break;
		case PT_MV_CLSID:
			ulCount = lpProp->Value.MVguid.cValues;
			break;
		default:
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}

		hr = HrWriteDWord(lpStream, ulCount);
	} else {
		ulCount = 1;
	}

	ulMVProp = 0;

	for(ulMVProp = 0; ulMVProp < ulCount; ulMVProp++) {
		switch(PROP_TYPE(lpProp->ulPropTag) &~ MV_FLAG) {
		case PT_I2:
			if(lpProp->ulPropTag & MV_FLAG)
				hr = HrWriteDWord(lpStream,lpProp->Value.MVi.lpi[ulMVProp]);
			else
				hr = HrWriteDWord(lpStream,lpProp->Value.i);
			break;
		case PT_LONG:
			if(lpProp->ulPropTag & MV_FLAG)
				hr = HrWriteDWord(lpStream,lpProp->Value.MVl.lpl[ulMVProp]);
			else
				hr = HrWriteDWord(lpStream,lpProp->Value.ul);
			break;
		case PT_BOOLEAN:
			hr = HrWriteDWord(lpStream, lpProp->Value.b);
			break;
		case PT_R4:
			if(lpProp->ulPropTag & MV_FLAG)
				hr = HrWriteData(lpStream,(char *)&lpProp->Value.MVflt.lpflt[ulMVProp], sizeof(float));
			else
				hr = HrWriteData(lpStream,(char *)&lpProp->Value.flt, sizeof(float));
			break;
		case PT_APPTIME:
			if(lpProp->ulPropTag & MV_FLAG)
				hr = HrWriteData(lpStream,(char *)&lpProp->Value.MVat.lpat[ulMVProp], sizeof(double));
			else
				hr = HrWriteData(lpStream,(char *)&lpProp->Value.at, sizeof(double));
			break;
		case PT_DOUBLE:
			if(lpProp->ulPropTag & MV_FLAG)
				hr = HrWriteData(lpStream,(char *)&lpProp->Value.MVdbl.lpdbl[ulMVProp], sizeof(double));
			else
				hr = HrWriteData(lpStream,(char *)&lpProp->Value.dbl, sizeof(double));
			break;
		case PT_CURRENCY:
			if(lpProp->ulPropTag & MV_FLAG) {
				hr = HrWriteDWord(lpStream, lpProp->Value.MVcur.lpcur[ulMVProp].Lo);
				if(hr != hrSuccess)
					goto exit;

				hr = HrWriteDWord(lpStream, lpProp->Value.MVcur.lpcur[ulMVProp].Hi);
				if(hr != hrSuccess)
					goto exit;
			} else {
				hr = HrWriteDWord(lpStream, lpProp->Value.cur.Lo);
				if(hr != hrSuccess)
					goto exit;

				hr = HrWriteDWord(lpStream, lpProp->Value.cur.Hi);
				if(hr != hrSuccess)
					goto exit;
			}
			break;
		case PT_SYSTIME:
			if(lpProp->ulPropTag & MV_FLAG) {
				hr = HrWriteDWord(lpStream, lpProp->Value.MVft.lpft[ulMVProp].dwLowDateTime);
				if(hr != hrSuccess)
					goto exit;

				hr = HrWriteDWord(lpStream, lpProp->Value.MVft.lpft[ulMVProp].dwHighDateTime);
				if(hr != hrSuccess)
					goto exit;
			} else {
				hr = HrWriteDWord(lpStream, lpProp->Value.ft.dwLowDateTime);
				if(hr != hrSuccess)
					goto exit;

				hr = HrWriteDWord(lpStream, lpProp->Value.ft.dwHighDateTime);
				if(hr != hrSuccess)
					goto exit;
			}
			break;
		case PT_I8:
			if(lpProp->ulPropTag & MV_FLAG) {
				hr = HrWriteDWord(lpStream, lpProp->Value.MVli.lpli[ulMVProp].LowPart);
				if(hr != hrSuccess)
					goto exit;

				hr = HrWriteDWord(lpStream, lpProp->Value.MVli.lpli[ulMVProp].HighPart);
				if(hr != hrSuccess)
					goto exit;
			} else {
				hr = HrWriteDWord(lpStream, lpProp->Value.li.LowPart);
				if(hr != hrSuccess)
					goto exit;

				hr = HrWriteDWord(lpStream, lpProp->Value.li.HighPart);
				if(hr != hrSuccess)
					goto exit;
			} 
			break;
		case PT_STRING8:
			if(lpProp->ulPropTag & MV_FLAG) {
				ulLen = strlen(lpProp->Value.MVszA.lppszA[ulMVProp])+1;

				hr = HrWriteDWord(lpStream, 1); // unknown why this is here
				if(hr != hrSuccess)
					goto exit;

				hr = HrWriteDWord(lpStream, ulLen);
				if(hr != hrSuccess)
					goto exit;

				hr = HrWriteData(lpStream, lpProp->Value.MVszA.lppszA[ulMVProp], ulLen);
				if(hr != hrSuccess)
					goto exit;
			} else {
				ulLen = strlen(lpProp->Value.lpszA)+1;

				hr = HrWriteDWord(lpStream, 1); // unknown why this is here
				if(hr != hrSuccess)
					goto exit;

				hr = HrWriteDWord(lpStream, ulLen);
				if(hr != hrSuccess)
					goto exit;

				hr = HrWriteData(lpStream, lpProp->Value.lpszA, ulLen);
				if(hr != hrSuccess)
					goto exit;
			}

			// Align to 4-byte boundary
			while(ulLen & 3) {
				hr = HrWriteByte(lpStream, 0);
				if (hr != hrSuccess)
					goto exit;
				ulLen++;
			}
			break;

		case PT_UNICODE:
			if(lpProp->ulPropTag & MV_FLAG) {
				ulLen = unicodelen(lpProp->Value.MVszW.lppszW[ulMVProp])*sizeof(WCHAR)+sizeof(WCHAR);

				hr = HrWriteDWord(lpStream, 1); // unknown why this is here
				if(hr != hrSuccess)
					goto exit;

				hr = HrWriteDWord(lpStream, ulLen);
				if(hr != hrSuccess)
					goto exit;

				hr = HrWriteData(lpStream, (char *)lpProp->Value.MVszW.lppszW[ulMVProp], ulLen);
				if(hr != hrSuccess)
					goto exit;
			} else {
				ulLen = unicodelen(lpProp->Value.lpszW)*sizeof(WCHAR)+sizeof(WCHAR);

				hr = HrWriteDWord(lpStream, 1); // unknown why this is here
				if(hr != hrSuccess)
					goto exit;

				hr = HrWriteDWord(lpStream, ulLen);
				if(hr != hrSuccess)
					goto exit;

				hr = HrWriteData(lpStream, (char *)lpProp->Value.lpszW, ulLen);
				if(hr != hrSuccess)
					goto exit;
			}

			// Align to 4-byte boundary
			while(ulLen & 3) {
				hr = HrWriteByte(lpStream, 0);
				if (hr != hrSuccess)
					goto exit;
				ulLen++;
			}
			break;

		case PT_BINARY:
			if(lpProp->ulPropTag & MV_FLAG) {
				ulLen = lpProp->Value.MVbin.lpbin[ulMVProp].cb;

				hr = HrWriteDWord(lpStream, 1); // unknown why this is here
				if(hr != hrSuccess)
					goto exit;

				hr = HrWriteDWord(lpStream, ulLen);
				if(hr != hrSuccess)
					goto exit;

				hr = HrWriteData(lpStream, (char *)lpProp->Value.MVbin.lpbin[ulMVProp].lpb, ulLen);
				if(hr != hrSuccess)
					goto exit;
			} else {
				ulLen = lpProp->Value.bin.cb;

				hr = HrWriteDWord(lpStream, 1); // unknown why this is here
				if(hr != hrSuccess)
					goto exit;

				hr = HrWriteDWord(lpStream, ulLen);
				if(hr != hrSuccess)
					goto exit;

				hr = HrWriteData(lpStream, (char *)lpProp->Value.bin.lpb, ulLen);
				if(hr != hrSuccess)
					goto exit;
			}

			// Align to 4-byte boundary
			while(ulLen & 3) {
				hr = HrWriteByte(lpStream, 0);
				if (hr != hrSuccess)
					goto exit;
				ulLen++;
			}
			break;

		// FIXME: PT_CLSID?

		default:
			hr = MAPI_E_INVALID_PARAMETER;
		}
	}

exit:
	if (lppNames)
		MAPIFreeBuffer(lppNames);

	return hr;
}

HRESULT ECTNEF::HrReadPropStream(char *lpBuffer, ULONG ulSize, std::list<SPropValue *> &proplist)
{
	ULONG ulRead = 0;
	ULONG ulProps = 0;
	LPSPropValue lpProp = NULL;
	HRESULT hr = hrSuccess;

	ulProps = *(ULONG *)lpBuffer;
	lpBuffer += 4;
	ulSize -= 4;

	// Loop through all the properties in the data and add them to our internal list
	while(ulProps) {
		hr = HrReadSingleProp(lpBuffer, ulSize, &ulRead, &lpProp);
		if(hr != hrSuccess)
			break;

		ulSize -= ulRead;
		lpBuffer += ulRead;

		proplist.push_back(lpProp);

		ulProps--;

		if(ulRead & 3) {
			// Skip padding
			lpBuffer += 4 - (ulRead & 3);
		}
	}

	return hr;
}

HRESULT ECTNEF::HrReadSingleProp(char *lpBuffer, ULONG ulSize, ULONG *lpulRead, LPSPropValue *lppProp)
{
	HRESULT hr = hrSuccess;
	ULONG ulPropTag = 0;
	ULONG ulLen = 0;
	ULONG ulOrigSize = ulSize;
	ULONG ulIsNameId = 0;
	ULONG ulCount = 0;
	ULONG ulMVProp = 0;
	LPSPropValue lpProp = NULL;
	GUID sGuid;
	MAPINAMEID sNameID;
	LPMAPINAMEID lpNameID = &sNameID;
	LPSPropTagArray lpPropTags = NULL;

	if(ulSize < 8) {
		return MAPI_E_NOT_FOUND;
	}

	ulPropTag = *(ULONG *)lpBuffer;

	lpBuffer += sizeof(ULONG);
	ulSize -= 4;

	hr = MAPIAllocateBuffer(sizeof(SPropValue), (void **)&lpProp);
	if(hr != hrSuccess)
		goto exit;

	if(PROP_ID(ulPropTag) >= 0x8000) {
		// Named property, first read GUID, then name/id
		if(ulSize < 24) {
			hr = MAPI_E_CORRUPT_DATA;
			goto exit;
		}
		memcpy(&sGuid, lpBuffer, sizeof(GUID));

		lpBuffer += sizeof(GUID);
		ulSize -= sizeof(GUID);

		ulIsNameId = *(ULONG *)lpBuffer;

		lpBuffer += 4;
		ulSize -= 4;

		if(ulIsNameId != 0) {
			// A string name follows
			ulLen = *(ULONG *)lpBuffer;

			lpBuffer += 4;
			ulSize -= 4;

			if(ulLen > ulSize) {
				hr = MAPI_E_CORRUPT_DATA;
				goto exit;
			}

			sNameID.ulKind = MNID_STRING;
			sNameID.Kind.lpwstrName = (WCHAR *)lpBuffer;

			lpBuffer += ulLen;
			ulSize -= ulLen;

			// Re-align
			lpBuffer += ulLen & 3 ? 4 - (ulLen & 3) : 0;
			ulSize -= ulLen & 3 ? 4 - (ulLen & 3) : 0;
		} else {
			sNameID.ulKind = MNID_ID;
			sNameID.Kind.lID = *(ULONG *)lpBuffer;

			lpBuffer += 4;
			ulSize -= 4;
		}

		sNameID.lpguid = &sGuid;

		hr = lpMessage->GetIDsFromNames(1, &lpNameID, MAPI_CREATE, &lpPropTags);
		if(hr != hrSuccess)
			goto exit;

		// Use the mapped ID, not the original ID. The original ID is discarded
		ulPropTag = PROP_TAG(PROP_TYPE(ulPropTag), PROP_ID(lpPropTags->aulPropTag[0]));
	}

	if(ulPropTag & MV_FLAG) {
		if(ulSize < 4) {
			hr = MAPI_E_CORRUPT_DATA;
			goto exit;
		}

		ulCount = *(ULONG *)lpBuffer;
		lpBuffer += 4;
		ulSize -= 4;
		
		switch(PROP_TYPE(ulPropTag)) {
		case PT_MV_I2:
			lpProp->Value.MVi.cValues = ulCount;
			hr = MAPIAllocateMore(ulCount * sizeof(unsigned short), lpProp, (void **)&lpProp->Value.MVi.lpi);
			break;
		case PT_MV_LONG:
			lpProp->Value.MVl.cValues = ulCount;
			hr = MAPIAllocateMore(ulCount * sizeof(ULONG), lpProp, (void **)&lpProp->Value.MVl.lpl);
			break;
		case PT_MV_R4:
			lpProp->Value.MVflt.cValues = ulCount;
			hr = MAPIAllocateMore(ulCount * sizeof(float), lpProp, (void **)&lpProp->Value.MVflt.lpflt);
			break;
		case PT_MV_APPTIME:
			lpProp->Value.MVat.cValues = ulCount;
			hr = MAPIAllocateMore(ulCount * sizeof(double), lpProp, (void **)&lpProp->Value.MVat.lpat);
			break;
		case PT_MV_DOUBLE:
			lpProp->Value.MVdbl.cValues = ulCount;
			hr = MAPIAllocateMore(ulCount * sizeof(double), lpProp, (void **)&lpProp->Value.MVdbl.lpdbl);
			break;
		case PT_MV_CURRENCY:
			lpProp->Value.MVcur.cValues = ulCount;
			hr = MAPIAllocateMore(ulCount * sizeof(CURRENCY), lpProp, (void **)&lpProp->Value.MVcur.lpcur);
			break;
		case PT_MV_SYSTIME:
			lpProp->Value.MVft.cValues = ulCount;
			hr = MAPIAllocateMore(ulCount * sizeof(FILETIME), lpProp, (void **)&lpProp->Value.MVft.lpft);
			break;
		case PT_MV_I8:
			lpProp->Value.MVli.cValues = ulCount;
			hr = MAPIAllocateMore(ulCount * sizeof(LARGE_INTEGER), lpProp, (void **)&lpProp->Value.MVli.lpli);
			break;
		case PT_MV_STRING8:
			lpProp->Value.MVszA.cValues = ulCount;
			hr = MAPIAllocateMore(ulCount * sizeof(char *), lpProp, (void **)&lpProp->Value.MVszA.lppszA);
			break;
		case PT_MV_UNICODE:
			lpProp->Value.MVszW.cValues = ulCount;
			hr = MAPIAllocateMore(ulCount * sizeof(WCHAR *), lpProp, (void **)&lpProp->Value.MVszW.lppszW);
			break;
		case PT_MV_BINARY:
			lpProp->Value.MVbin.cValues = ulCount;
			hr = MAPIAllocateMore(ulCount * sizeof(SBinary), lpProp, (void **)&lpProp->Value.MVbin.lpbin);
			break;
		case PT_MV_CLSID:
			lpProp->Value.MVguid.cValues = ulCount;
			hr = MAPIAllocateMore(ulCount * sizeof(GUID), lpProp, (void **)&lpProp->Value.MVi.lpi);
			break;
		default:
			hr = MAPI_E_INVALID_PARAMETER;
			goto exit;
		}
	} else {
		ulCount = 1;
	}

	if(hr != hrSuccess)
		goto exit;

	lpProp->ulPropTag = ulPropTag;

	for(ulMVProp = 0; ulMVProp < ulCount; ulMVProp++) {
		switch(PROP_TYPE(ulPropTag) & ~MV_FLAG) {
		case PT_I2:
			if(ulPropTag & MV_FLAG)
				lpProp->Value.MVi.lpi[ulMVProp] = *(unsigned short *)lpBuffer;
			else
				lpProp->Value.i = *(unsigned short *)lpBuffer;

			lpBuffer += 4;
			ulSize -= 4;
			break;
		case PT_LONG:
			if(ulPropTag & MV_FLAG)
				lpProp->Value.MVl.lpl[ulMVProp] = *(ULONG *)lpBuffer;
			else
				lpProp->Value.ul = *(ULONG *)lpBuffer;

			lpBuffer += 4;
			ulSize -= 4;
			break;
		case PT_BOOLEAN:
			lpProp->Value.b = *(BOOL *)lpBuffer;
			lpBuffer += 4;
			ulSize -= 4;
			break;
		case PT_R4:
			if(ulPropTag & MV_FLAG)
				lpProp->Value.MVflt.lpflt[ulMVProp] = *(float *)lpBuffer;
			else
				lpProp->Value.flt = *(float *)lpBuffer;

			lpBuffer += 4;
			ulSize -= 4;
			break;
		case PT_APPTIME:
			if(ulSize < 8) {
				hr = MAPI_E_CORRUPT_DATA;
				goto exit;
			}
			if(ulPropTag & MV_FLAG)
				lpProp->Value.MVat.lpat[ulMVProp] = *(double *)lpBuffer;
			else
				lpProp->Value.at = *(double *)lpBuffer;

			lpBuffer += 8;
			ulSize -= 8;
			break;
		case PT_DOUBLE:
			if(ulSize < 8) {
				hr = MAPI_E_CORRUPT_DATA;
				goto exit;
			}
			if(ulPropTag & MV_FLAG)
				lpProp->Value.MVdbl.lpdbl[ulMVProp] = *(double *)lpBuffer;
			else
				lpProp->Value.dbl = *(double *)lpBuffer;

			lpBuffer += 8;
			ulSize -= 8;
			break;
		case PT_CURRENCY:
			if(ulSize < 8) {
				hr = MAPI_E_CORRUPT_DATA;
				goto exit;
			}
			if(ulPropTag & MV_FLAG) {
				lpProp->Value.MVcur.lpcur[ulMVProp].Lo = *(ULONG *)lpBuffer;
				lpProp->Value.MVcur.lpcur[ulMVProp].Hi = *(ULONG *)(lpBuffer+4);
			} else {
				lpProp->Value.cur.Lo = *(ULONG *)lpBuffer;
				lpProp->Value.cur.Hi = *(ULONG *)lpBuffer+4;
			}

			lpBuffer += 8;
			ulSize -= 8;
			break;
		case PT_SYSTIME:
			if(ulSize < 8) {
				hr = MAPI_E_CORRUPT_DATA;
				goto exit;
			}
			if(ulPropTag & MV_FLAG) {
				lpProp->Value.MVft.lpft[ulMVProp].dwLowDateTime = *(ULONG *)lpBuffer;
				lpProp->Value.MVft.lpft[ulMVProp].dwHighDateTime = *(ULONG *)(lpBuffer+4);
			} else {
				lpProp->Value.ft.dwLowDateTime = *(ULONG *)lpBuffer;
				lpProp->Value.ft.dwHighDateTime = *(ULONG *)(lpBuffer+4);
			}
			lpBuffer += 8;
			ulSize -= 8;
			break;
		case PT_I8:
			if(ulSize < 8) {
				hr = MAPI_E_CORRUPT_DATA;
				goto exit;
			}
			if(ulPropTag & MV_FLAG) {
				lpProp->Value.MVli.lpli[ulMVProp].LowPart = *(ULONG *)lpBuffer;
				lpProp->Value.MVli.lpli[ulMVProp].HighPart = *(ULONG *)(lpBuffer+4);
			} else {
				lpProp->Value.li.LowPart = *(ULONG *)lpBuffer;
				lpProp->Value.li.HighPart = *(ULONG *)(lpBuffer+4);
			} 

			lpBuffer += 8;
			ulSize -= 8;
			break;
		case PT_STRING8:
			if(ulSize < 8) {
				hr = MAPI_E_CORRUPT_DATA;
				goto exit;
			}

			lpBuffer += 4; // Skip next 4 bytes, they are always '1'
			ulSize -= 4;

			ulLen = *(ULONG *)lpBuffer;

			lpBuffer += 4; 
			ulSize -= 4;

			if(ulSize < ulLen) {
				hr = MAPI_E_CORRUPT_DATA;
				goto exit;
			}
			if(ulPropTag & MV_FLAG) {
				hr = MAPIAllocateMore(ulLen+1, lpProp, (void **)&lpProp->Value.MVszA.lppszA[ulMVProp]);
				if(hr != hrSuccess)
					goto exit;

				memcpy(lpProp->Value.MVszA.lppszA[ulMVProp], lpBuffer, ulLen);
				lpProp->Value.MVszA.lppszA[ulMVProp][ulLen] = 0; // should be terminated anyway but we terminte it just to be sure
			} else {
				hr = MAPIAllocateMore(ulLen+1, lpProp, (void **)&lpProp->Value.lpszA);
				if(hr != hrSuccess)
					goto exit;

				memcpy(lpProp->Value.lpszA, lpBuffer, ulLen);
				lpProp->Value.lpszA[ulLen] = 0; // should be terminated anyway but we terminte it just to be sure
			}

			lpBuffer += ulLen;
			ulSize -= ulLen;

			// Re-align
			lpBuffer += ulLen & 3 ? 4 - (ulLen & 3) : 0;
			ulSize -= ulLen & 3 ? 4 - (ulLen & 3) : 0;
			break;

		case PT_UNICODE:
			if(ulSize < 8) {
				hr = MAPI_E_CORRUPT_DATA;
				goto exit;
			}

			lpBuffer += 4; // Skip next 4 bytes, they are always '1'
			ulSize -= 4;

			ulLen = *(ULONG *)lpBuffer;	// Assumes 'len' in file is BYTES, not chars

			lpBuffer += 4;
			ulSize -= 4;

			if(ulSize < ulLen) {
				hr = MAPI_E_CORRUPT_DATA;
				goto exit;
			}

			// adds extra WCHAR for absolute string termination
			if(ulPropTag & MV_FLAG) {
				hr = MAPIAllocateMore(ulLen+sizeof(WCHAR), lpProp, (void **)&lpProp->Value.MVszW.lppszW[ulMVProp]);

				if(hr != hrSuccess)
					goto exit;

				memcpy(lpProp->Value.MVszW.lppszW[ulMVProp], lpBuffer, ulLen);
				lpProp->Value.MVszW.lppszW[ulMVProp][ulLen/sizeof(WCHAR)] = 0;
			} else {
				hr = MAPIAllocateMore(ulLen+sizeof(WCHAR), lpProp, (void **)&lpProp->Value.lpszW);
				if(hr != hrSuccess)
					goto exit;

				memcpy(lpProp->Value.lpszW, lpBuffer, ulLen);
				lpProp->Value.lpszW[ulLen/sizeof(WCHAR)] = 0;
			}

			lpBuffer += ulLen;
			ulSize -= ulLen;

			// Re-align
			lpBuffer += ulLen & 3 ? 4 - (ulLen & 3) : 0;
			ulSize -= ulLen & 3 ? 4 - (ulLen & 3) : 0;
			break;

		case PT_OBJECT:			// PST sends PT_OBJECT data. Treat as PT_BINARY
		case PT_BINARY:
			if(ulSize < 8) {
				hr = MAPI_E_CORRUPT_DATA;
				goto exit;
			}
			lpBuffer += 4;	// Skip next 4 bytes, it's always '1' (ULONG)
			ulSize -= 4;

			ulLen = *(ULONG *)lpBuffer;

			lpBuffer += 4;
			ulSize -= 4;

			if (PROP_TYPE(ulPropTag) == PT_OBJECT) {
				// TODO: IMessage guid [ 0x00020307 C000 0000 0000 0000 00 00 00 46 ]
				lpBuffer += 16;
				ulSize -= 16;
				ulLen -= 16;
			}

			if(ulSize < ulLen) {
				hr = MAPI_E_CORRUPT_DATA;
				goto exit;
			}

			if(ulPropTag & MV_FLAG) {
				hr = MAPIAllocateMore(ulLen, lpProp, (void **)&lpProp->Value.MVbin.lpbin[ulMVProp].lpb);
				if(hr != hrSuccess)
					goto exit;

				memcpy(lpProp->Value.MVbin.lpbin[ulMVProp].lpb, lpBuffer, ulLen);
				lpProp->Value.MVbin.lpbin[ulMVProp].cb = ulLen;				
			} else {
				hr = MAPIAllocateMore(ulLen, lpProp, (void **)&lpProp->Value.bin.lpb);
				if(hr != hrSuccess)
					goto exit;

				memcpy(lpProp->Value.bin.lpb, lpBuffer, ulLen);
				lpProp->Value.bin.cb = ulLen;
			}

			lpBuffer += ulLen;
			ulSize -= ulLen;

			// Re-align
			lpBuffer += ulLen & 3 ? 4 - (ulLen & 3) : 0;
			ulSize -= ulLen & 3 ? 4 - (ulLen & 3) : 0;
			break;

		// FIXME: PT_CLSID?

		default:
			hr = MAPI_E_INVALID_PARAMETER;
			break;
		}
	}

	*lpulRead = ulOrigSize - ulSize;
	*lppProp = lpProp;

exit:
	if(hr != hrSuccess) {
		if(lpProp)
			MAPIFreeBuffer(lpProp);
	}

	if(lpPropTags)
		MAPIFreeBuffer(lpPropTags);

	return hr;
}


HRESULT ECTNEF::SetProps(ULONG cValues, LPSPropValue lpProps)
{
	unsigned int i = 0;

	for(i=0; i<cValues; i++) {
		lstProps.push_back(&lpProps[i]);
	}

	return hrSuccess;
}

HRESULT ECTNEF::Finish()
{
	HRESULT hr = hrSuccess;
	std::list<SPropValue *>::iterator iterProps;
	IStream *lpPropStream = NULL;
	STATSTG sStat;
	ULONG ulChecksum;
	LARGE_INTEGER zero = {{0,0}};
	// attachment vars
	ULONG ulAttachNum;
	LPATTACH lpAttach = NULL;
	LPSTREAM lpAttStream = NULL;
	LPMESSAGE lpAttMessage = NULL;
	IStream *lpSubStream = NULL;
	std::list<tnefattachment*>::iterator iterAttach;

	if(ulFlags == TNEF_DECODE) {
		// Write properties to message
		for(iterProps = lstProps.begin(); iterProps != lstProps.end(); iterProps++) {

			if(PROP_ID((*iterProps)->ulPropTag) == PROP_ID(PR_MESSAGE_CLASS) ||
				!FPropExists(lpMessage, (*iterProps)->ulPropTag) ||
				PROP_ID((*iterProps)->ulPropTag) == PROP_ID(PR_RTF_COMPRESSED))
			{
				lpMessage->SetProps(1, *iterProps, NULL);
			}
			// else, Property already exists, do *not* overwrite it

		}
		// Add all found attachments to message
		for (iterAttach = lstAttachments.begin(); iterAttach != lstAttachments.end(); iterAttach++) {
			bool has_obj = false;

			hr = lpMessage->CreateAttach(NULL, 0, &ulAttachNum, &lpAttach);
			if (hr != hrSuccess)
				goto exit;
			for (iterProps = (*iterAttach)->lstProps.begin(); iterProps != (*iterAttach)->lstProps.end(); iterProps++) {
				// must not set PR_ATTACH_NUM by ourselves
				if (PROP_ID((*iterProps)->ulPropTag) == PROP_ID(PR_ATTACH_NUM))
					continue;

				if ((*iterProps)->ulPropTag != PR_ATTACH_DATA_OBJ) {
					hr = lpAttach->SetProps(1, *iterProps, NULL);
					if (hr != hrSuccess)
						goto exit;
				} else {
					// message in PT_OBJECT, was saved in Value.bin
					hr = CreateStreamOnHGlobal(NULL, TRUE, &lpSubStream);
					if (hr != hrSuccess)
						goto exit;

					hr = lpSubStream->Write((*iterProps)->Value.bin.lpb, (*iterProps)->Value.bin.cb, NULL);
					if (hr != hrSuccess)
						goto exit;

					hr = lpSubStream->Seek(zero, STREAM_SEEK_SET, NULL);
					if(hr != hrSuccess)
						goto exit;

					hr = lpAttach->OpenProperty(PR_ATTACH_DATA_OBJ, &IID_IMessage, 0, MAPI_CREATE|MAPI_MODIFY, (LPUNKNOWN *)&lpAttMessage);
					if(hr != hrSuccess)
						goto exit;

					ECTNEF SubTNEF(TNEF_DECODE, lpAttMessage, lpSubStream);
					hr = SubTNEF.ExtractProps(TNEF_PROP_EXCLUDE, NULL);
					if (hr != hrSuccess)
						goto exit;

					hr = SubTNEF.Finish();
					if (hr != hrSuccess)
						goto exit;

					hr = lpAttMessage->SaveChanges(0);
					if (hr != hrSuccess)
						goto exit;

					has_obj = true;
				}
			}
			if (!has_obj) {
				hr = lpAttach->OpenProperty(PR_ATTACH_DATA_BIN, &IID_IStream, STGM_WRITE|STGM_TRANSACTED, MAPI_CREATE|MAPI_MODIFY, (LPUNKNOWN *)&lpAttStream);
				if (hr != hrSuccess)
					goto exit;

				hr = lpAttStream->Write((*iterAttach)->data,(*iterAttach)->size,NULL);
				if (hr != hrSuccess)
					goto exit;
				hr = lpAttStream->Commit(0);
				if (hr != hrSuccess)
					goto exit;
				lpAttStream->Release();
				lpAttStream = NULL;
			}

			hr = lpAttach->SaveChanges(0);
			if (hr != hrSuccess)
				goto exit;
			lpAttach->Release();
			lpAttach = NULL;
		}
	} else if(ulFlags == TNEF_ENCODE) {
		// Write properties to stream
		hr = HrWriteDWord(lpStream, TNEF_SIGNATURE);
		if(hr != hrSuccess)
			goto exit;

		hr = HrWriteWord(lpStream, 0); // Write Key
		if(hr != hrSuccess)
			goto exit;

		hr = HrWriteByte(lpStream, 1); // Write component (always 1 ?)
		if(hr != hrSuccess)
			goto exit;

		hr = CreateStreamOnHGlobal(NULL, TRUE, &lpPropStream);
		if(hr != hrSuccess)
			goto exit;

		hr = HrWritePropStream(lpPropStream);
		if(hr != hrSuccess)
			goto exit;

		hr = lpPropStream->Stat(&sStat, STATFLAG_NONAME);
		if(hr != hrSuccess)
			goto exit;

		hr = HrWriteDWord(lpStream, 0x00069003);
		if(hr != hrSuccess)
			goto exit;

		hr = HrWriteDWord(lpStream, sStat.cbSize.LowPart); // Write size
		if(hr != hrSuccess)
			goto exit;

		hr = lpPropStream->Seek(zero, STREAM_SEEK_SET, NULL);
		if(hr != hrSuccess)
			goto exit;

		hr = lpPropStream->CopyTo(lpStream, sStat.cbSize, NULL, NULL); // Write data
		if(hr != hrSuccess)
			goto exit;

		hr = lpPropStream->Seek(zero, STREAM_SEEK_SET, NULL);
		if(hr != hrSuccess)
			goto exit;

		hr = HrGetChecksum(lpPropStream, &ulChecksum);
		if(hr != hrSuccess)
			goto exit;

		hr = HrWriteWord(lpStream, (unsigned short)ulChecksum); // Write checksum
		if(hr != hrSuccess)
			goto exit;
	}

exit:
	if (lpSubStream)
		lpSubStream->Release();

	if (lpAttMessage)
		lpAttMessage->Release();

	if (lpAttStream)
		lpAttStream->Release();

	if (lpAttach)
		lpAttach->Release();

	if(lpPropStream)
		lpPropStream->Release();

	return hr;
}

HRESULT ECTNEF::HrReadDWord(IStream *lpStream, ULONG *ulData)
{
	HRESULT hr = hrSuccess;
	ULONG ulRead = 0;

	hr = lpStream->Read(ulData, sizeof(unsigned int), &ulRead);
	if(hr != hrSuccess)
		goto exit;

	if(ulRead != sizeof(unsigned int)) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

exit:
	return hr;
}

HRESULT ECTNEF::HrReadWord(IStream *lpStream, unsigned short *ulData)
{
	HRESULT hr = hrSuccess;
	ULONG ulRead = 0;

	hr = lpStream->Read(ulData, sizeof(unsigned short), &ulRead);
	if(hr != hrSuccess)
		goto exit;

	if(ulRead != sizeof(unsigned short)) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

exit:
	return hr;
}

HRESULT ECTNEF::HrReadByte(IStream *lpStream, unsigned char *ulData)
{
	HRESULT hr = hrSuccess;
	ULONG ulRead = 0;

	hr = lpStream->Read(ulData, sizeof(char), &ulRead);
	if(hr != hrSuccess)
		goto exit;

	if(ulRead != sizeof(char)) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

exit:
	return hr;
}

HRESULT ECTNEF::HrReadData(IStream *lpStream, char *lpData, ULONG ulLen)
{
	HRESULT hr = hrSuccess;
	ULONG ulRead = 0;
	ULONG ulToRead = 0;

	while(ulLen) {
		ulToRead = ulLen > 4096 ? 4096 : ulLen;

		hr = lpStream->Read(lpData, ulToRead, &ulRead);
		if(hr != hrSuccess)
			goto exit;

		if(ulRead != ulToRead) {
			hr = MAPI_E_NOT_FOUND;
			goto exit;
		}

		ulLen -= ulRead;
		lpData += ulRead;
	}

exit:
	return hr;
}

HRESULT ECTNEF::HrWriteDWord(IStream *lpStream, ULONG ulData)
{
	HRESULT hr = hrSuccess;
	ULONG ulWritten = 0;

	hr = lpStream->Write(&ulData, sizeof(unsigned int), &ulWritten);
	if(hr != hrSuccess)
		goto exit;

	if(ulWritten != sizeof(unsigned int)) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

exit:
	return hr;
}

HRESULT ECTNEF::HrWriteWord(IStream *lpStream, unsigned short ulData)
{
	HRESULT hr = hrSuccess;
	ULONG ulWritten = 0;

	hr = lpStream->Write(&ulData, sizeof(unsigned short), &ulWritten);
	if(hr != hrSuccess)
		goto exit;

	if(ulWritten != sizeof(unsigned short)) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

exit:
	return hr;
}

HRESULT ECTNEF::HrWriteByte(IStream *lpStream, unsigned char ulData)
{
	HRESULT hr = hrSuccess;
	ULONG ulWritten = 0;

	hr = lpStream->Write(&ulData, sizeof(unsigned char), &ulWritten);
	if(hr != hrSuccess)
		goto exit;

	if(ulWritten != sizeof(unsigned char)) {
		hr = MAPI_E_NOT_FOUND;
		goto exit;
	}

exit:
	return hr;
}

HRESULT ECTNEF::HrWriteData(IStream *lpStream, char *data, ULONG ulLen)
{
	HRESULT hr = hrSuccess;
	ULONG ulWritten = 0;

	while(ulLen > 0) {
		hr = lpStream->Write(data, ulLen > 4096 ? 4096 : ulLen, &ulWritten);
		if(hr != hrSuccess)
			goto exit;

		ulLen -= ulWritten;
		data += ulWritten;
	}

exit:
	return hr;
}

// TNEF uses the rather stupid checksum of adding all the bytes in the stream. 
// Was TNEF coded by an intern or something ??
HRESULT ECTNEF::HrGetChecksum(IStream *lpStream, ULONG *lpulChecksum)
{
	HRESULT hr = hrSuccess;
	ULONG ulChecksum = 0;
	IStream *lpClone = NULL;
	LARGE_INTEGER zero = {{0,0}};
	ULONG ulRead = 0;
	unsigned char buffer[4096];
	unsigned int i = 0;

	hr = lpStream->Clone(&lpClone);
	if(hr != hrSuccess)
		goto exit;

	hr = lpClone->Seek(zero, STREAM_SEEK_SET, NULL);
	if(hr != hrSuccess)
		goto exit;

	while(TRUE) {
		hr = lpClone->Read(buffer, 4096, &ulRead);
		if(hr != hrSuccess)
			goto exit;

		if(ulRead == 0)
			break;

		for(i=0;i<ulRead;i++) {
			ulChecksum += buffer[i];
		}
	}

	*lpulChecksum = ulChecksum;

exit:
	if(lpClone)
		lpClone->Release();

	return hr;
}
